import re

from bidsschematools import rules

from ..types import Namespace


def test_entity_rule(schema_obj):
    # Simple
    rule = Namespace.build(
        {
            "datatypes": ["anat"],
            "entities": {"subject": "required", "session": "optional"},
            "suffixes": ["T1w"],
            "extensions": [".nii"],
        }
    )
    nii_rule = rules._entity_rule(rule, schema_obj)
    assert nii_rule == {
        "regex": (
            r"sub-(?P<subject>[0-9a-zA-Z]+)/"
            r"(?:ses-(?P<session>[0-9a-zA-Z]+)/)?"
            r"(?P<datatype>anat)/"
            r"(?(subject)sub-(?P=subject)_)"
            r"(?(session)ses-(?P=session)_)"
            r"(?P<suffix>T1w)"
            r"(?P<extension>\.nii)\Z"
        ),
        "mandatory": False,
    }

    assert re.match(nii_rule["regex"], "sub-01/anat/sub-01_T1w.nii")
    assert re.match(nii_rule["regex"], "sub-01/ses-01/anat/sub-01_ses-01_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01/anat/sub-02_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01/sub-01_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01/ses-01/anat/sub-01_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01/anat/sub-01_ses-01_T1w.nii")
    assert not re.match(nii_rule["regex"], "sub-01/ses-01/anat/sub-01_ses-02_T1w.nii")

    # Sidecar entities are optional
    rule = Namespace.build(
        {
            "datatypes": ["anat", ""],
            "entities": {"subject": "optional", "session": "optional"},
            "suffixes": ["T1w"],
            "extensions": [".json"],
        }
    )
    json_rule = rules._entity_rule(rule, schema_obj)
    assert json_rule == {
        "regex": (
            r"(?:sub-(?P<subject>[0-9a-zA-Z]+)/)?"
            r"(?:ses-(?P<session>[0-9a-zA-Z]+)/)?"
            r"(?:(?P<datatype>anat)/)?"
            r"(?(subject)sub-(?P=subject)_)"
            r"(?(session)ses-(?P=session)_)"
            r"(?P<suffix>T1w)"
            r"(?P<extension>\.json)\Z"
        ),
        "mandatory": False,
    }
    assert re.match(json_rule["regex"], "sub-01/anat/sub-01_T1w.json")
    assert re.match(json_rule["regex"], "sub-01/sub-01_T1w.json")
    assert re.match(json_rule["regex"], "T1w.json")
    assert re.match(json_rule["regex"], "sub-01/ses-01/anat/sub-01_ses-01_T1w.json")
    assert re.match(json_rule["regex"], "sub-01/ses-01/sub-01_ses-01_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01/anat/sub-02_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01_T1w.json")
    assert not re.match(json_rule["regex"], "ses-01_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01/ses-01/anat/sub-01_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01/anat/sub-01_ses-01_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01/ses-01/ses-01_T1w.json")
    assert not re.match(json_rule["regex"], "sub-01/ses-01/anat/sub-01_ses-02_T1w.json")


def test_split_inheritance_rules():
    rule = {
        "datatypes": ["anat"],
        "entities": {"subject": "required", "session": "optional"},
        "suffixes": ["T1w"],
        "extensions": [".nii", ".json"],
    }

    main, sidecar = rules._split_inheritance_rules(rule)
    assert main == {
        "datatypes": ["anat"],
        "entities": {"subject": "required", "session": "optional"},
        "suffixes": ["T1w"],
        "extensions": [".nii"],
    }
    assert sidecar == {
        "datatypes": ["", "anat"],
        "entities": {"subject": "optional", "session": "optional"},
        "suffixes": ["T1w"],
        "extensions": [".json"],
    }

    # Can't split again
    (main2,) = rules._split_inheritance_rules(main)
    assert main2 == {
        "datatypes": ["anat"],
        "entities": {"subject": "required", "session": "optional"},
        "suffixes": ["T1w"],
        "extensions": [".nii"],
    }


def test_stem_rule():
    rule = Namespace.build({"stem": "README", "level": "required", "extensions": ["", ".md"]})
    assert rules._stem_rule(rule) == {
        "regex": r"(?P<stem>(?s:README))(?P<extension>|\.md)\Z",
        "mandatory": True,
    }

    rule = Namespace.build(
        {"stem": "participants", "level": "optional", "extensions": [".tsv", ".json"]}
    )
    assert rules._stem_rule(rule) == {
        "regex": r"(?P<stem>(?s:participants))(?P<extension>\.tsv|\.json)\Z",
        "mandatory": False,
    }

    # Wildcard stem, with datatype
    rule = Namespace.build(
        {
            "stem": "*",
            "datatypes": ["phenotype"],
            "level": "optional",
            "extensions": [".tsv", ".json"],
        }
    )
    assert rules._stem_rule(rule) == {
        "regex": r"(?P<datatype>phenotype)/(?P<stem>(?s:.*))(?P<extension>\.tsv|\.json)\Z",
        "mandatory": False,
    }


def test_path_rule():
    rule = Namespace.build({"path": "dataset_description.json", "level": "required"})
    assert rules._path_rule(rule) == {
        "regex": r"(?P<path>dataset_description\.json)(?:/.*)?\Z",
        "mandatory": True,
    }

    rule = Namespace.build({"path": "LICENSE", "level": "optional"})
    assert rules._path_rule(rule) == {"regex": r"(?P<path>LICENSE)(?:/.*)?\Z", "mandatory": False}


def test_regexify_all():
    schema_all, _ = rules.regexify_all()

    # Check if expected keys are present in all entries
    for entry in schema_all:
        assert "regex" in entry
        assert "mandatory" in entry



"""A partial implementation of schema-based validation in Python."""

import datetime
import fnmatch
import json
import os
import re
from copy import deepcopy
from pathlib import Path

import bidsschematools as bst
import bidsschematools.rules
import bidsschematools.schema
import bidsschematools.types
import bidsschematools.utils

lgr = bst.utils.get_logger()

VALIDATOR_SCHEMA_COMPATIBILITY_LEVEL = "minor"


def _bids_schema_versioncheck(schema_dir, compatibility=VALIDATOR_SCHEMA_COMPATIBILITY_LEVEL):
    """
    Use the `SCHEMA_VERSION` descriptor file to determine whether a selected schema directory is
    compatible with the validator.

    Parameters
    ----------
    schema_dir : str
        A string which specifies a path.
    compatibility : "major" or "minor", optional
        Compatibility range of the validator. Major means "M.*" are considered compatible, minor
        means "M.m.*" are considered compatible.

    Returns
    -------
    bool:
        Whether the schema is compatible with the validator.
    """

    if compatibility not in ("major", "minor"):
        raise ValueError("Schema compatibility needs to be set to either “major” or “minor”.")

    if isinstance(schema_dir, str):
        schema_dir = Path(schema_dir)

    schema_version_file = schema_dir.joinpath("SCHEMA_VERSION")
    try:
        schema_version = schema_version_file.read_text().strip()
    except FileNotFoundError:
        lgr.warning(
            "The selected schema directory, `%s`, does not contain a SCHEMA_VERSION file. "
            "Cannot ascertain compatibility. Attempting to query the BIDS reference for "
            "compatible versions.",
            schema_dir,
        )
    else:
        nparts = 1 if compatibility == "major" else 2
        return schema_version.split(".", nparts)[:-1] == bst.__version__.split(".", nparts)[:-1]
        lgr.warning(
            "The selected schema `%s`, has a schema version (`%s`) which is "
            "incompatible with the validator. Attempting to query the BIDS reference "
            "for compatible versions.",
            schema_dir,
            schema_version,
        )
    return False


def _find_bids_root(in_paths, accept_non_bids_dir):
    """
    Return BIDS root for a list of paths.
    Raise error if more than one root is found and, optionally, if none is found.
    """

    dataset_descriptions = []
    for in_path in in_paths:
        in_path = os.path.abspath(os.path.expanduser(in_path))
        dataset_description = _find_dataset_description(in_path)
        if dataset_description and dataset_description not in dataset_descriptions:
            dataset_descriptions.append(dataset_description)
    if len(dataset_descriptions) > 1:
        raise ValueError(
            f"You have selected files belonging to {len(dataset_descriptions)} "
            "different datasets. Please run the validator once per dataset."
        )
    elif len(dataset_descriptions) == 0:
        if accept_non_bids_dir:
            lgr.warning(
                "None of the files in the input list are part of a BIDS dataset. Proceeding."
            )
            return ""
        else:
            raise ValueError(
                "None of the files in the input list are part of a BIDS dataset. Aborting."
            )
    return os.path.dirname(dataset_descriptions[0])


def _find_dataset_description(my_path):
    my_path = Path(my_path)
    for path in (my_path, *my_path.parents):
        candidate = path / "dataset_description.json"
        if candidate.is_file():
            return candidate
    return None


def _get_directory_suffixes(my_schema):
    """Query schema for suffixes which identify directory entities.

    Parameters
    ----------
    my_schema : dict
        Nested directory as produced by `bidsschematools.schema.load_schema()`.

    Returns
    -------
    list of str
        Directory pseudofile suffixes excluding trailing slashes.

    Notes
    -----
    * Yes this seems super-awkward to do explicitly, after all, the trailing slash is
        already in so it should automagically work, but no:
        - Subdirectory names need to be dynamically excluded from validation input.
        - Backslash directory delimiters are still in use, which is regrettable.
    """
    return [
        ext.value[:-1]
        for ext in my_schema.objects.extensions.values()
        if len(ext.value) > 1 and ext.value.endswith("/")
    ]


def _get_paths(
    bids_paths,
    pseudofile_suffixes=None,
    dummy_paths=False,
    exclude_files=None,
):
    """
    Get all paths from a list of directories, excluding hidden subdirectories from distribution.

    Parameters
    ----------
    bids_paths : list or str
        Directories from which to get paths, may also contain file paths, which will remain
        unchanged.
    pseudofile_suffixes : list of str
        Directory suffixes prompting the validation of the directory name and limiting further
        directory walk.
    dummy_paths : bool, optional
        Whether to accept path strings which do not correspond to either files or directories.
    exclude_files : list, optional
        Files to exclude from listing.
        Dot files (`.*`) do not need to be explicitly listed, as these are excluded by default.

    Notes
    -----
    * Deduplicate paths (if input dirs are subsets of other input dirs), might best be done at the
        very end. This is only relevant for poor usage (e.g. passing parent *and* child
        directories), and has thus far not caused problems, but it would be good to brace for
        that.
    * The `dataset_description.json` and `.bidsignore` handling should be split out of the main
        loop, since we now have BIDS root detection called before file detection. This is
        non-critical however, since the topdown `os.walk` makes sure top-level files are detected
        first.
    """

    if exclude_files is None:
        exclude_files = []
    if pseudofile_suffixes is None:
        pseudofile_suffixes = []

    path_list = []
    bidsignore_list = []
    bids_root_found = False
    for bids_path in bids_paths:
        if not dummy_paths:
            bids_path = os.path.abspath(os.path.expanduser(bids_path))
        if os.path.isdir(bids_path):
            for root, dirs, file_names in os.walk(bids_path, topdown=True):
                if "dataset_description.json" in file_names:
                    if bids_root_found:
                        # No nested BIDS.
                        dirs[:] = []
                        file_names[:] = []
                    else:
                        try:
                            with open(os.path.join(root, ".bidsignore")) as f:
                                bidsignore_list = f.read().splitlines()
                        except FileNotFoundError:
                            pass
                        bids_root_found = True
                if root.endswith(tuple(pseudofile_suffixes)):
                    # Add the directory name to the validation paths list.
                    path_list.append(Path(root).as_posix() + "/")
                    # Do not index the contents of the directory.
                    dirs[:] = []
                    file_names[:] = []
                if os.path.basename(root).startswith("."):
                    dirs[:] = []
                    file_names[:] = []
                for file_name in file_names:
                    if file_name in exclude_files or file_name.startswith("."):
                        continue
                    if bidsignore_list:
                        ignored = False
                        for ignore_expression in bidsignore_list:
                            ignored = _bidsignore_check(ignore_expression, file_name, root)
                            if ignored:
                                break
                        if ignored:
                            continue
                    file_path = os.path.join(root, file_name)
                    path_list.append(Path(file_path).as_posix())
        elif os.path.isfile(bids_path) or dummy_paths:
            path_list.append(Path(bids_path).as_posix())
        else:
            raise FileNotFoundError(
                f"The input path `{bids_path}` could not be located. If this is a string "
                "intended for path validation which does not correspond to an actual "
                "path, please set the `dummy_paths` parameter to True."
            )

    return path_list


def _bidsignore_check(ignore_expression, file_name, file_root):
    """
    Check whether a file is set to be ignored as per `.bidsignore`

    Parameters
    ----------
    ignore_expression : str
        A string following git-wildcard conventions (including `**/`).
    file_name : str
        A string which represents a filename.
    file_root : str
        The directory containing the file specifiled in `file_name`.

    Returns
    -------
    bool:
        Whether the file should be ignored.

    Notes
    -----
    * We cannot use `glob` since that would preempt working with theoretical or non-local
        paths, therefore we use `fnmatch`.
    * `fnmatch` does not support `**/` matching as that is an optional convention from e.g.
        globstar and git, and not part of the standard Unix shell. As we formalize `.bidsignore`
        we may choose drop it, since we already treat simple filenames as to-be-ignored in
        all directories, and with BIDS having only up to 4 hierarchical levels, the utility of
        other usage is limited and expansion to `..*/*..` would only mean at maximum a duplication
        of entries.
    """

    if ignore_expression.startswith("**/"):
        ignore_expression = ignore_expression.lstrip("**/")
    elif any(i in ignore_expression for i in ["/", "\\"]):
        file_name = os.path.join(file_root, file_name)

    return fnmatch.fnmatch(file_name, ignore_expression)


def log_errors(validation_result):
    """
    Raise errors for validation result.

    Parameters
    ----------
    validation_result : dict
        A dictionary as returned by `validate_all()` with keys including "schema_tracking",
        "path_tracking", "path_listing", and, optionally "itemwise".
        The "itemwise" value, if present, should be a list of dictionaries, with keys including
        "path", "regex", and "match".
    """
    total_file_count = len(validation_result["path_listing"])
    validated_files_count = total_file_count - len(validation_result["path_tracking"])
    errorless = True
    for i in validation_result["path_tracking"]:
        lgr.warning("The `%s` file was not matched by any regex schema entry.", i)
        errorless = False
    if validated_files_count == 0:
        lgr.error("No valid BIDS files were found.")
        errorless = False
    else:
        # No use reporting this separately if no BIDS files were found
        for entry in validation_result["schema_tracking"]:
            if entry["mandatory"]:
                lgr.error(
                    "The `%s` regex pattern file required by BIDS was not found.",
                    entry["regex"],
                )
                errorless = False
    if errorless:
        lgr.info("SUCCESS: All files are BIDS valid and no BIDS-required files are missing.")


def select_schema_path(
    bids_version=None,
    bids_root=None,
    bids_reference_root=None,
):
    """
    Select schema directory, according to a priority logic whereby the schema path is
    either:

    (1) a concatenation of `bids_reference_root` and `bids_version`, if the latter is
        specified, and the BIDS version schema is compatible with the validator,
    (2) a concatenation of `bids_reference_root` the detected version specification
        inside the BIDS root directory, if such a directory is provided and the BIDS version
        schema is compatible with the validator.
    (3) `None`, expanded to the bundled schema.

    Parameters
    ----------
    bids_root : str or None, optional
        The path to the BIDS root for the paths to be validated.
    bids_reference_root : str, optional
        Path where schema versions are stored, and which contains directories named exactly
        according to the respective schema version, e.g. "1.7.0".
    bids_version : str or None, optional
        BIDS version desired for validation.
        If empty, the `dataset_description.json` fie will be queried for the dataset schema
        version.


    Returns
    -------
    str
        A string which is a path to the selected schema directory.

    Notes
    -----
    * This is a purely aspirational function, and is preempted by logic inside
        `bst.validator.validate_bids()`, and further contingent on better schema stability and
        ongoing work in: https://github.com/bids-standard/bids-schema
    * The default `bids_reference_root` value is based on the FHS and ideally should be enforced.
        Alternatively this could be handled by an environment variable, though that also requires
        enforcement on the package distribution side.
    """

    if bids_reference_root is None:
        lgr.warning("No BIDS reference root provided.")
        return None

    bids_reference_root = os.path.abspath(os.path.expanduser(bids_reference_root))

    schema_dir = False
    if bids_root and not bids_version:
        dataset_description = os.path.join(bids_root, "dataset_description.json")
        with open(dataset_description) as f:
            try:
                dataset_info = json.load(f)
            except json.decoder.JSONDecodeError:
                lgr.error(
                    "The `%s` file cannot be loaded. Please check whether it is valid JSON.",
                    dataset_description,
                )
            else:
                try:
                    bids_version = dataset_info["BIDSVersion"]
                except KeyError:
                    lgr.warning("BIDSVersion is not specified in `dataset_description.json`.")
    if bids_version:
        schema_dir = os.path.join(bids_reference_root, bids_version)
        if _bids_schema_versioncheck(schema_dir):
            return schema_dir

    try:
        for schema_dir_candidate in os.listdir(bids_reference_root):
            schema_dir = os.path.join(bids_reference_root, schema_dir_candidate)
            if _bids_schema_versioncheck(schema_dir):
                return schema_dir
    except FileNotFoundError:
        pass
    lgr.warning(
        "No suitable schema could be found in the BIDS reference root (`%s`).",
        bids_reference_root,
    )
    return None


def validate_all(
    paths_list,
    regex_schema,
):
    """
    Validate `bids_paths` based on a `regex_schema` dictionary list, including regexes.

    Parameters
    ----------
    paths_list : list or str
        A string pointing to a BIDS directory for which paths should be validated, or a list
        of strings pointing to individual files or subdirectories which *all* reside within
        one and only one BIDS directory root (i.e. nested datasets should be validated
        separately).
    regex_schema : list of dict
        A list of dictionaries as generated by `regexify_all()`.

    Returns
    -------
    results : dict
        A dictionary reporting the target files for validation, the unmatched files and unmatched
        regexes, and optionally the itemwise comparison results.
        Keys include "schema_tracking", "path_tracking", "path_listing", "match_listing", and
        optionally "itemwise"

    Notes
    -----
    * Multi-source validation could be accomplished by distributing the resulting tracking_schema
        dictionary and further eroding it.
    * Currently only entities are captured in named groups, edit `load_top_level()` to name other
        groups as well.
    """

    tracking_paths = deepcopy(paths_list)
    tracking_schema = []
    itemwise_results = []
    matched = False
    match_listing = []
    for target_path in paths_list:
        lgr.debug("Checking file `%s`.", target_path)
        lgr.debug("Trying file types:")
        for regex_entry in regex_schema:
            target_regex = r"(?:.*/)?" + regex_entry["regex"]
            # We need to record the actual expressions we query.
            _regex_entry = deepcopy(regex_entry)
            _regex_entry.update({"regex": target_regex})
            lgr.debug("\t* `%s`, with pattern: `%s`", target_path, target_regex)
            matched = re.match(target_regex, target_path)
            itemwise_result = {}
            itemwise_result["path"] = target_path
            itemwise_result["regex"] = target_regex
            if matched:
                lgr.debug("Match identified.")
                itemwise_result["match"] = True
                itemwise_results.append(itemwise_result)
                break
            itemwise_result["match"] = False
            itemwise_results.append(itemwise_result)
        if matched:
            tracking_paths.remove(target_path)
            # Might be fragile since it relies on where the loop broke:
            if not regex_entry["mandatory"]:
                tracking_schema.append(_regex_entry)
            match_entry = matched.groupdict()
            match_entry["path"] = target_path
            match_listing.append(match_entry)
        else:
            tracking_schema.append(_regex_entry)
            lgr.debug(
                "The `%s` file could not be matched to any regex schema entry.",
                target_path,
            )
    results = {}
    results["itemwise"] = itemwise_results
    results["schema_tracking"] = tracking_schema
    results["schema_listing"] = regex_schema
    results["path_tracking"] = tracking_paths
    results["path_listing"] = paths_list
    results["match_listing"] = match_listing

    return results


def write_report(
    validation_result,
    report_path="~/.cache/bidsschematools/validator-report_{datetime}-{pid}.log",
    datetime_format="%Y%m%d%H%M%SZ",
):
    """Write a human-readable report based on the validation result.

    Parameters
    ----------
    validation_result : dict
        A dictionary as returned by `validate_all()` with keys including "schema_tracking",
        "path_tracking", "path_listing", and, optionally "itemwise".
        The "itemwise" value, if present, should be a list of dictionaries, with keys including
        "path", "regex", and "match".
    report_path : str, optional
        A path under which the report is to be saved, `datetime`, and `pid`
        are available as variables for string formatting, and will be expanded to the
        current datetime (as per the `datetime_format` parameter)
        and process ID, respectively.
    datetime_format : str, optional
        A datetime format, optionally used for the report path.

    Notes
    -----
    * Not using f-strings in order to prevent arbitrary code execution.
    """

    report_path = report_path.format(
        datetime=datetime.datetime.utcnow().strftime(datetime_format),
        pid=os.getpid(),
    )
    report_path = os.path.abspath(os.path.expanduser(report_path))
    report_dir = os.path.dirname(report_path)
    try:
        os.makedirs(report_dir)
    except OSError:
        pass

    total_file_count = len(validation_result["path_listing"])
    validated_files_count = total_file_count - len(validation_result["path_tracking"])
    with open(report_path, "w") as f:
        try:
            for comparison in validation_result["itemwise"]:
                if comparison["match"]:
                    comparison_result = "A MATCH"
                else:
                    comparison_result = "no match"
                f.write(
                    f"- Comparing the `{comparison['path']}` path to the `{comparison['regex']}` "
                    f"pattern resulted in {comparison_result}.\n"
                )
        except KeyError:
            pass
        f.write(
            f"\nSUMMARY:\n{validated_files_count} out of {total_file_count} files were "
            "successfully validated, using the following regular expressions:"
        )
        for regex_entry in validation_result["schema_listing"]:
            f.write(f"\n\t- `{regex_entry['regex']}`")
        f.write("\n")
        if len(validation_result["path_tracking"]) > 0:
            f.write("The following files were not matched by any regex schema entry:")
            f.write("\n\t* `")
            f.write("`\n\t* `".join(validation_result["path_tracking"]))
        else:
            f.write("All files were matched by a regex schema entry.")
        if len(validation_result["schema_tracking"]) > 0:
            f.write("\nThe following mandatory regex schema entries did not match any files:")
            f.write("\n")
            for entry in validation_result["schema_tracking"]:
                if entry["mandatory"]:
                    f.write(f"\t** `{entry['regex']}`\n")
        else:
            f.write("All mandatory BIDS files were found.\n")
        f.close()
    lgr.info("BIDS validation log written to %s", report_path)


def validate_bids(
    in_paths,
    dummy_paths=False,
    bids_reference_root=None,
    schema_path=None,
    bids_version=None,
    report_path=False,
    suppress_errors=False,
    accept_non_bids_dir=False,
    exclude_files=None,
):
    """
    Validate paths according to BIDS schema.

    Parameters
    ----------
    in_paths : str or list of str
        Paths which to validate, may be individual files or directories.
    dummy_paths : bool, optional
        Whether to accept path strings which do not correspond to either files or directories.
    bids_reference_root : str, optional
        Path where schema versions are stored, and which contains directories named exactly
        according to the respective schema version, e.g. "1.7.0".
        Currently this is untested.
    bids_version : str or None, optional
        Version of BIDS schema, or path to schema. This supersedes the specification detected in
        `dataset_description.json` and is itself superseded if `schema_path` is specified.
    schema_path : str or None, optional
        If a path is given, this will be expanded and used directly, ignoring all other BIDS
        version specification logic. This is not relative to `bids_reference_root`.
    report_path : bool or str, optional
        If `True` a log will be written using the standard output path of `.write_report()`.
        If string, the string will be used as the output path.
        If the variable evaluates as False, no log will be written.
    accept_non_bids_dir : bool, optional
    exclude_files : str, optional
        Files which will not be indexed for validation, use this if your data is in an archive
        standard which requires the presence of archive-specific files (e.g. DANDI requiring
        `dandiset.yaml`).
        Dot files (`.*`) do not need to be explicitly listed, as these are excluded by default.

    Returns
    -------
    results : dict
        A dictionary reporting the target files for validation, the unmatched files and unmatched
        regexes, and optionally the itemwise comparison results.
        Keys include "schema_tracking", "path_tracking", "path_listing", "match_listing", and
        optionally "itemwise"

    Examples
    --------

    ::

        from bidsschematools import validator
        bids_paths = '~/.data2/datalad/000026/noncompliant'
        validator.validate_bids(bids_paths)

    Notes
    -----
    * Needs to account for inheritance principle, probably somewhere deeper in the logic, might be
        as simple as pattern parsing and multiplying patterns to which inheritance applies.
        https://github.com/bids-standard/bids-specification/pull/969#issuecomment-1132119492
    """

    if exclude_files is None:
        exclude_files = []

    if isinstance(in_paths, str):
        in_paths = [in_paths]

    # Are we dealing with real paths?
    if dummy_paths:
        bids_root = None
    else:
        bids_root = _find_bids_root(in_paths, accept_non_bids_dir)

    # Select schema path:
    if not schema_path:
        schema_path = select_schema_path(
            bids_version,
            bids_root,
            bids_reference_root=bids_reference_root,
        )

    regex_schema, my_schema = bst.rules.regexify_all(schema_path)
    pseudofile_suffixes = _get_directory_suffixes(my_schema)

    # Get list of all paths since inputs can be directories.
    bids_paths = _get_paths(
        in_paths,
        dummy_paths=dummy_paths,
        pseudofile_suffixes=pseudofile_suffixes,
        exclude_files=exclude_files,
    )

    # Go!
    validation_result = validate_all(
        bids_paths,
        regex_schema,
    )

    # Record schema version.
    validation_result["bids_version"] = my_schema["bids_version"]

    log_errors(validation_result)

    if report_path:
        if isinstance(report_path, str):
            write_report(validation_result, report_path=report_path)
        else:
            write_report(validation_result)

    return validation_result



import os
import shutil

import pytest

from bidsschematools.conftest import BIDS_ERROR_SELECTION, BIDS_SELECTION
from bidsschematools.validator import select_schema_path, validate_bids

from ..data import load
from .data import load_test_data


def test_inheritance_examples():
    correct_inheritance = [
        "/lala/sub-01/ses-test/sub-01_ses-test_task-sometask_bold.json",
        "/lala/sub-01/sub-01_task-sometask_bold.json",
        "/lala/task-sometask_bold.json",
    ]
    incorrect_inheritance = [
        "/lala/sub-01/sub-01_ses-test_task-sometask_bold.json",
        "/lala/ses-test_task-sometask.json",
    ]

    result = validate_bids(
        correct_inheritance + incorrect_inheritance,
        dummy_paths=True,
    )

    assert result["path_tracking"] == incorrect_inheritance


def test_regression_examples():
    """Tests that failed on pybids when switching to bst-regex-based validation"""
    examples = [
        "/sub-01/ses-ses/sub-01_dwi.bval",  # redundant dir /ses-ses/
        "/sub-01/01_dwi.bvec",  # missed subject suffix
        "/sub-01/sub_dwi.json",  # missed subject id
        "/sub-01/sub-01_23_run-01_dwi.bval",  # wrong _23_
        "/sub-01/sub-01_run-01_dwi.vec",  # wrong extension
        "/sub-01/sub-01_run-01_dwi.jsn",  # wrong extension
        "/sub-01/sub-01_acq_dwi.bval",  # missed suffix value
        "/sub-01/sub-01_acq-23-singleband_dwi.bvec",  # redundant -23-
        "/sub-01/anat/sub-01_acq-singleband_dwi.json",  # redundant /anat/
        "/sub-01/sub-01_recrod-record_acq-singleband_run-01_dwi.bval",  # redundant record-record_
        "/sub_01/sub-01_acq-singleband_run-01_dwi.bvec",  # wrong /sub_01/
        "/sub-01/sub-01_acq-singleband__run-01_dwi.json",  # wrong __
        "/sub-01/ses-test/sub-01_ses_test_dwi.bval",  # wrong ses_test
        "/sub-01/ses-test/sb-01_ses-test_dwi.bvec",  # wrong sb-01
        "/sub-01/ses-test/sub-01_ses-test_dw.json",  # wrong modality
        "/sub-01/ses-test/sub-01_ses-test_run-01_dwi.val",  # wrong extension
        "/sub-01/ses-test/sub-01_run-01_dwi.bvec",  # missed session in the filename
        # This validator adds a .*/ to the regex, so this will be a false negative
        # If I cared about this validator, I would dig into it, but it doesn't seem worth it.
        # -cjm 2024.08.14
        # "/sub-01/ses-test/ses-test_run-01_dwi.json",  # missed subject in the filename
        "/sub-01/ses-test/sub-01_ses-test_acq-singleband.bval",  # missed modality
        "/sub-01/ses-test/sub-01_ses-test_acq-singleband_dwi",  # missed extension
        "/ses-test/sub-01/sub-01_ses-test_acq-singleband_dwi.json",  # wrong dirs order
        "/sub-01/ses-test/sub-02_ses-test_acq-singleband_run-01_dwi.bval",  # wrong sub id
        "/sub-01/sub-01_ses-test_acq-singleband_run-01_dwi.bvec",  # ses dir missed
        "/ses-test/sub-01_ses-test_acq-singleband_run-01_dwi.json",  # sub id dir missed
    ]

    result = validate_bids(examples, dummy_paths=True)
    assert result["path_tracking"] == examples


def test_write_report(tmp_path):
    from bidsschematools.validator import write_report

    validation_result = {}

    validation_result["schema_tracking"] = [
        {
            "regex": ".*?/sub-(?P<subject>[0-9a-zA-Z]+)/"
            "(|ses-(?P<session>[0-9a-zA-Z]+)/)anat/sub-(?P=subject)"
            "(|_ses-(?P=session))(|_acq-(?P<acquisition>[0-9a-zA-Z]+))"
            "(|_ce-(?P<ceagent>[0-9a-zA-Z]+))"
            "(|_rec-(?P<reconstruction>[0-9a-zA-Z]+))"
            "(|_run-(?P<run>[0-9a-zA-Z]+))"
            "(|_part-(?P<part>(mag|phase|real|imag)))"
            "_(T1w|T2w|PDw|T2starw|FLAIR|inplaneT1|inplaneT2|PDT2|angio|T2star)"
            "\\.(nii.gz|nii|json)$",
            "mandatory": False,
        }
    ]
    validation_result["schema_listing"] = [
        {
            "regex": ".*?/sub-(?P<subject>[0-9a-zA-Z]+)/"
            "(|ses-(?P<session>[0-9a-zA-Z]+)/)anat/sub-(?P=subject)"
            "(|_ses-(?P=session))(|_acq-(?P<acquisition>[0-9a-zA-Z]+))"
            "(|_ce-(?P<ceagent>[0-9a-zA-Z]+))"
            "(|_rec-(?P<reconstruction>[0-9a-zA-Z]+))"
            "(|_run-(?P<run>[0-9a-zA-Z]+))"
            "(|_part-(?P<part>(mag|phase|real|imag)))"
            "_(T1w|T2w|PDw|T2starw|FLAIR|inplaneT1|inplaneT2|PDT2|angio|T2star)"
            "\\.(nii.gz|nii|json)$",
            "mandatory": False,
        }
    ]
    validation_result["path_tracking"] = [
        "/home/chymera/.data2/datalad/000026/"
        "noncompliant/sub-EXC022/anat/sub-EXC022_ses-MRI_flip-1_VFA.nii.gz"
    ]
    validation_result["path_listing"] = [
        "/home/chymera/.data2/datalad/000026/"
        "noncompliant/sub-EXC022/anat/sub-EXC022_ses-MRI_flip-1_VFA.nii.gz"
    ]

    report_path = tmp_path / "output_bids_validator_xs_write.log"
    write_report(validation_result, report_path=str(report_path))

    expected_report_path = load_test_data("expected_bids_validator_xs_write.log")
    assert report_path.read_text() == expected_report_path.read_text()


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
@pytest.mark.parametrize("dataset", BIDS_SELECTION)
def test_bids_datasets(bids_examples, tmp_path, dataset):
    # Validate per dataset:
    target = os.path.join(bids_examples, dataset)
    result = validate_bids(
        target,
    )
    # Have all files been validated?
    assert len(result["path_tracking"]) == 0


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
def test_validate_bids(bids_examples, tmp_path):
    # Create input for file list based validation
    selected_dir = os.path.join(bids_examples, BIDS_SELECTION[0])
    selected_paths = []
    for root, dirs, files in os.walk(selected_dir, topdown=False):
        for f in files:
            selected_path = os.path.join(root, f)
            selected_paths.append(selected_path)
    # Does version fallback work?
    result = validate_bids(selected_paths, schema_path=False)
    # Does default log path specification work?
    result = validate_bids(selected_paths, report_path=True)

    # Does custom log path specification work?
    result = validate_bids(
        selected_paths,
        report_path=os.path.join(tmp_path, "test_bids.log"),
    )
    # Have all files been validated?
    assert len(result["path_tracking"]) == 0

    # Is the schema version recorded correctly?
    schema_path = load.readable("schema")
    expected_version = schema_path.joinpath("BIDS_VERSION").read_text().rstrip()
    assert result["bids_version"] == expected_version


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
def test_broken_json_dataset(bids_examples, tmp_path):
    """Perhaps this can be integrated into
    https://github.com/bids-standard/bids-error-examples ."""

    dataset = "asl003"
    dataset_path = os.path.join(bids_examples, dataset)
    dataset_json = os.path.join(dataset_path, "dataset_description.json")

    broken_json = load_test_data("broken_dataset_description.json")
    shutil.copyfile(broken_json, dataset_json)

    # No assert, will simply raise JSON reader error if not catching it properly.
    _ = validate_bids(
        dataset_path,
        report_path=True,
    )


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
def test_exclude_files(bids_examples, tmp_path):
    from bidsschematools.validator import validate_bids

    dataset = "asl003"
    dataset_reference = os.path.join(bids_examples, dataset)
    tmp_path = str(tmp_path)
    shutil.copytree(dataset_reference, tmp_path, dirs_exist_ok=True)

    # Create non-BIDS non-dotfile
    archive_file_name = "dandiset.yaml"
    archive_file_path = os.path.join(tmp_path, archive_file_name)
    with open(archive_file_path, "w") as f:
        f.write(" \n")

    # Does it fail, as it should (more like a failsafe assertion)
    result = validate_bids(tmp_path)
    assert len(result["path_tracking"]) == 1

    # Does the parameter work?
    result = validate_bids(tmp_path, exclude_files=[archive_file_name])
    assert len(result["path_tracking"]) == 0


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
def test_accept_non_bids_dir(bids_examples, tmp_path):
    from bidsschematools.validator import validate_bids

    dataset = "asl003"
    dataset_reference = os.path.join(bids_examples, dataset)
    tmp_path = str(tmp_path)
    shutil.copytree(dataset_reference, tmp_path, dirs_exist_ok=True)

    # remove `dataset_description.json`
    os.remove(os.path.join(tmp_path, "dataset_description.json"))

    # Does it fail, as it should (more like a failsafe assertion)
    with pytest.raises(
        ValueError,
        match="None of the files in the input list are part of a BIDS dataset. Aborting.",
    ):
        _ = validate_bids(tmp_path)

    # Does the parameter work?
    result = validate_bids(tmp_path, accept_non_bids_dir=True)
    assert len(result["path_tracking"]) == 0


@pytest.mark.skipif(
    os.environ.get("SCHEMACODE_TESTS_NONETWORK") is not None,
    reason="no network",
)
@pytest.mark.parametrize("dataset", BIDS_ERROR_SELECTION)
def test_error_datasets(bids_error_examples, dataset):
    target = os.path.join(bids_error_examples, dataset)
    result = validate_bids(
        target,
        report_path=True,
    )
    # Are there non-validated files?
    assert len(result["path_tracking"]) != 0


def test_gitdir(bids_examples, tmp_path):
    """Maybe better handled in example data?"""

    selected_dir = os.path.join(bids_examples, BIDS_SELECTION[0])
    tmp_path = str(tmp_path)
    shutil.copytree(selected_dir, tmp_path, dirs_exist_ok=True)

    os.makedirs(os.path.join(tmp_path, ".git"))
    with open(os.path.join(tmp_path, ".git", "config"), "w") as temp_file:
        temp_file.write("")
    result = validate_bids(tmp_path)
    assert len(result["path_tracking"]) == 0


def test_select_schema_path(bids_examples, tmp_path):
    dataset = "asl003"
    dataset_path = os.path.join(bids_examples, dataset)

    # Does fallback to None work without any `raise`?
    schema_path = select_schema_path(dataset_path)
    assert schema_path is None


def test_bids_schema_versioncheck(monkeypatch, schema_dir):
    """Test incompatible version."""
    import bidsschematools as bst

    schema_dir = bst.data.load.readable("schema")
    assert bst.validator._bids_schema_versioncheck(schema_dir)
    monkeypatch.setattr(bst, "__version__", "99.99.99")
    assert not bst.validator._bids_schema_versioncheck(schema_dir)



"""A Python package for working with the BIDS schema.

.. autodata:: __version__
.. autodata:: __bids_version__
"""

__version__: str
__bids_version__: str

__all__ = ("__version__", "__bids_version__")


def __getattr__(attr: str) -> str:
    """Lazily load the schema version and BIDS version from the filesystem."""
    from typing import TypeVar

    from .data import load

    T = TypeVar("T")

    def document(obj: T, docstring: str) -> T:
        tp = type(obj)
        return type(tp.__name__, (tp,), {"__doc__": docstring})(obj)

    versions = {
        "__version__": (
            "schema/SCHEMA_VERSION",
            "schema_version",
            "Schema version",
        ),
        "__bids_version__": (
            "schema/BIDS_VERSION",
            "bids_version",
            "BIDS specification version",
        ),
    }

    if attr in versions:
        dir_path, schema_path, docstring = versions[attr]

        # Fast path if the schema directory is present (editable mode)
        if (version_file := load.readable(dir_path)).is_file():
            version = version_file.read_text().strip()
        else:
            # If version files are absent, the schema.json has been packaged.
            # If we're going to read it, we might as well cache it with load_schema().
            from .schema import load_schema

            version = load_schema()[schema_path]
        globals()[attr] = document(version, docstring)
        return globals()[attr]

    raise AttributeError(f"module {__spec__.name!r} has no attribute {attr!r}")



from bidsschematools.types import context as ctx
from bidsschematools.types import protocols as p


def test_import():
    "Verify that the module contains the generated content."
    assert ctx.__doc__ is not None
    assert ctx.__doc__.splitlines()[0].startswith("BIDS validation context")
    assert "Context" in ctx.__all__

    assert isinstance(ctx.Context, type)


def test_assignability() -> None:
    """Verify that dataclass values can be assigned to variables annotated with protocols.

    For pytest, this just checks instantiability.
    Running mypy with bst installed should check assignability,
    for example, with::

        uvx --with=. mypy tests
    """
    subjects: p.Subjects = ctx.Subjects([])
    subjects = ctx.Subjects([], [], [])

    dataset: p.Dataset = ctx.Dataset(
        dataset_description={},
        tree={},
        ignored=[],
        datatypes=[],
        modalities=[],
        subjects=subjects,
    )

    magnitude: p.Magnitude = ctx.Magnitude("path")
    magnitude1: p.Magnitude1 = ctx.Magnitude1("path")
    m0scan: p.M0scan = ctx.M0scan("path")
    bval: p.Bval = ctx.Bval("path", 5, 1, [0, 0, 0, 0, 0])
    channels: p.Channels = ctx.Channels("path")
    channels = ctx.Channels("path", ["TYPE"], ["1"], ["SHORT"])
    events: p.Events = ctx.Events("path", ["0.0", "3.0"])
    bvec: p.Bvec = ctx.Bvec("path", 5, 3)
    coordsystem: p.Coordsystem = ctx.Coordsystem("path")
    aslcontext: p.Aslcontext = ctx.Aslcontext("path", 2, ["label", "control"])
    associations: p.Associations = ctx.Associations()
    associations = ctx.Associations(
        magnitude=magnitude,
        magnitude1=magnitude1,
        m0scan=m0scan,
        bval=bval,
        channels=channels,
        events=events,
        bvec=bvec,
        coordsystem=coordsystem,
        aslcontext=aslcontext,
    )

    gzip: p.Gzip = ctx.Gzip(0)
    gzip = ctx.Gzip(0, "filename", "comment")

    sessions: p.Sessions = ctx.Sessions([])
    sessions = ctx.Sessions([], [], [])
    subject: p.Subject = ctx.Subject(sessions)

    tiff: p.Tiff = ctx.Tiff(0x4D4D)

    dim_info: p.DimInfo = ctx.DimInfo(1, 2, 3)
    xyzt_units: p.XyztUnits = ctx.XyztUnits("mm", "sec")
    nifti_header: p.NiftiHeader = ctx.NiftiHeader(
        dim_info=dim_info,
        dim=[4, 64, 64, 48, 100, 1, 1, 1],
        pixdim=[1.0, 1.0, 1.0, 1.0, 1.5, 1.0, 1.0, 1.0],
        shape=(64, 64, 48, 100),
        voxel_sizes=(1.0, 1.0, 1.0, 1.5),
        xyzt_units=xyzt_units,
        qform_code=1,
        sform_code=1,
    )

    ome: p.Ome = ctx.Ome()
    ome = ctx.Ome(
        PhysicalSizeX=10,
        PhysicalSizeY=10,
        PhysicalSizeZ=10,
        PhysicalSizeXUnit="um",
        PhysicalSizeYUnit="um",
        PhysicalSizeZUnit="um",
    )

    context: p.Context = ctx.Context(
        schema={},
        dataset=dataset,
        path="path",
        size=0,
        sidecar={},
        associations=associations,
    )
    context = ctx.Context(
        schema={},
        dataset=dataset,
        subject=subject,
        path="path",
        modality="modality",
        datatype="datatype",
        entities={},
        suffix="suffix",
        extension=".ext",
        size=0,
        sidecar={},
        associations=associations,
        gzip=gzip,
        tiff=tiff,
        ome=ome,
        nifti_header=nifti_header,
        columns={},
        json={},
    )
    assert context.schema == {}  # Use the variable to pacify linters



"""Module containing data files, including the schema source files

.. autofunction:: load

.. automethod:: load.readable

.. automethod:: load.as_path

.. automethod:: load.cached
"""

from acres import Loader

load = Loader(__spec__.name)



import logging
import tempfile
from collections.abc import Generator
from pathlib import Path
from subprocess import run

import pytest

from . import data

lgr = logging.getLogger()

# This selects a subset of the bids-examples collection to run the test suite on.
# Generally it's best to avoid adding large datasets to this list, but ideally a
# good proportion modalities would be represented, as well as datasets exemplifying
# tricky edge-cases, such as directory pseudo-files.
BIDS_SELECTION = [
    "asl003",  # anat, perf, _asl, _T1w
    "eeg_cbm",  # eeg
    "hcp_example_bids",  # anat, fmap
    "micr_SEMzarr",  # micr, SEM, OME-ZARR
    "micr_SPIM",  # micr, SPIM, .ome.tif
    "pet003",  # pet, anat
    "qmri_tb1tfl",  # fmap, _TB1TFL
    "qmri_vfa",  # derivatives
    "ds000248",  # .bidsignore
    "fnirs_automaticity",  # phenotypic
]
# Errors are described in the README of the respective datasets:
# https://github.com/bids-standard/bids-error-examples
BIDS_ERROR_SELECTION = [
    "invalid_asl003",
    "invalid_pet001",
]


@pytest.fixture(scope="session")
def tests_data_dir():
    try:
        this_file = Path(__file__)
    except NameError:
        return None

    data_dir = this_file.parent.parent.parent / "tests" / "data"

    if data_dir.exists():
        return data_dir


def get_gitrepo_fixture(url, whitelist):
    @pytest.fixture(scope="session")
    def fixture(tests_data_dir):
        if tests_data_dir is None:
            pytest.skip("No test data directory found; probably in an installed package")
        archive_name = url.rsplit("/", 1)[-1]
        archive_dir = tests_data_dir / archive_name
        if archive_dir.is_dir():
            lgr.info(
                f"Found static testdata archive under `{archive_dir}`. "
                "Not downloading latest data from version control."
            )
            yield archive_dir
        else:
            lgr.info(
                "No static testdata available under `%s`. "
                "Attempting to fetch live data from version control.",
                archive_dir,
            )
            with tempfile.TemporaryDirectory() as path:
                lgr.debug("Cloning %r into %r", url, path)
                runout = run(
                    [
                        "git",
                        "clone",
                        "--depth=1",
                        "--filter=blob:none",
                        "--sparse",
                        url,
                        path,
                    ],
                    capture_output=True,
                )
                if runout.returncode:
                    raise RuntimeError(f"Failed to clone {url} into {path}")
                # cwd specification is VERY important, not only to achieve the correct
                # effects, but also to avoid dropping files from your repository if you
                # were to run `git sparse-checkout` inside the software repo.
                _ = run(["git", "sparse-checkout", "init", "--cone"], cwd=path)
                _ = run(["git", "sparse-checkout", "set"] + whitelist, cwd=path)
                yield path

    return fixture


@pytest.fixture(scope="session")
def schema_dir() -> Generator[str, None, None]:
    """Path to the schema housed in the bids-specification repo."""
    with data.load.as_path("schema") as schema_path:
        if not schema_path.exists():
            pytest.skip("No schema found; probably in an installed package")
        yield str(schema_path)


@pytest.fixture(scope="session")
def schema_obj():
    """Schema object."""
    from bidsschematools import schema

    return schema.load_schema()


bids_examples = get_gitrepo_fixture(
    "https://github.com/bids-standard/bids-examples",
    whitelist=BIDS_SELECTION,
)
bids_error_examples = get_gitrepo_fixture(
    "https://github.com/bids-standard/bids-error-examples",
    whitelist=BIDS_ERROR_SELECTION,
)



"""Loader for dynamically-generated context.

This module is accessed when the package is installed in editable mode.
This source code should not be found in the installed package.
"""

from ..schema import load_schema
from ._generator import generate_module

schema = load_schema()
exec(generate_module(schema, "protocol"), globals())



"""Functions for rendering portions of the schema as text."""

from __future__ import annotations

import pandas as pd
from tabulate import tabulate

from bidsschematools.render import utils
from bidsschematools.schema import BIDSSchemaError, Namespace, filter_schema
from bidsschematools.utils import get_logger

lgr = get_logger()

# Remember to add extension (.html or .md) to the paths when using them.
ENTITIES_PATH = "SPEC_ROOT/appendices/entities"
GLOSSARY_PATH = "SPEC_ROOT/glossary"


def _make_object_table(
    subschema,
    field_info,
    table_type=None,
    src_path=None,
    tablefmt="github",
    n_values_to_combine=15,
):
    """Make a generic table describing objects in the schema.

    This does shared work between describing metadata fields and subobjects in the schema.

    Parameters
    ----------
    subschema : Namespace or dict
        Subset of the overall schema, including only the target object definitions.
    field_info : dict
        Additional information about each row in the table to be added to schema information.
        Keys should match "name" entries in ``subschema``.
        Values should be either a string (in which case, it's requirement information) or
        a tuple (in which case, the first entry is requirement info and the second is
        information to be added to the object's description).
    table_type : str
        The name of the field type. For example, "metadata".
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).
    n_values_to_combine : int, optional
        When there are many valid values for a given object,
        instead of listing each one in the table,
        link to the associated glossary entry.
    """
    # Use the "name" field in the table, to allow for filenames to not match "names".
    first_column = "**Column name**" if table_type == "columns" else "**Key name**"
    df = pd.DataFrame(
        index=field_info.keys(),
        columns=[
            first_column,
            "**Requirement Level**",
            "**Data type**",
            "**Description**",
        ],
    )
    element_type = {
        "metadata": "field",
        "subobject": "subfield",
        "columns": "column",
    }.get(table_type)

    for element, field in subschema.items():
        field_name = field["name"]
        # NOTE: Link to the glossary entry,
        # except for subobjects (if table_type) and
        # "additional columns" (field_name.startswith("**"))
        if table_type and not field_name.startswith("**"):
            field_name = f"[{field_name}]({GLOSSARY_PATH}.md#objects.{table_type}.{element})"

        # Grab any requirement level information and text to be added to the description string
        requirement_info, description_addendum = field_info[element]["table_info"]
        requirement_info = utils.normalize_requirements(requirement_info)
        requirement_info = requirement_info.replace(
            "DEPRECATED",
            "[DEPRECATED](SPEC_ROOT/common-principles.md#definitions)",
        )

        type_string = utils.resolve_metadata_type(field)

        description = utils.normalize_requirements(
            f"{field['description']} {description_addendum}".strip()
        )

        # Append a list of valid values, if provided, to the description.
        # If there are a lot of valid values, this will add a link to the description linking to
        # the associated glossary entry.
        levels = subschema[element].get("enum", []) or subschema[element].get(
            "definition", {}
        ).get("Levels", [])
        if len(levels) >= n_values_to_combine:
            glossary_entry = f"{GLOSSARY_PATH}.md#objects.{table_type}.{element}"
            valid_values_str = (
                f"For a list of valid values for this {element_type}, see the "
                f"[associated glossary entry]({glossary_entry})."
            )
        else:
            # Try to add info about valid values
            valid_values_str = utils.describe_valid_values(subschema[element])

        if valid_values_str:
            description += "\n\n\n\n" + valid_values_str

        # Add entry to DataFrame
        df.loc[element] = [
            field_name,
            utils.normalize_breaks(requirement_info),
            type_string,
            utils.normalize_breaks(description),
        ]

    df = df.set_index(first_column, drop=True)
    df.index.name = first_column

    # Print it as markdown
    table_str = tabulate(df, headers="keys", tablefmt=tablefmt)

    # Spec internal links need to be replaced
    table_str = table_str.replace("SPEC_ROOT", utils.get_relpath(src_path))

    return table_str


def _make_table_from_rule(
    schema: Namespace,
    table_type: str,
    table_name: str | list[str],
    src_path: str | None = None,
    tablefmt: str = "github",
):
    """Create a table for one or more rules.

    If ``table_type`` is "columns", only one table may be provided.

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    table_type : {"metadata", "columns"}
        The type of table. Either "metadata" or "columns".
    table_name : str or list of str
        Qualified name(s) in schema.rules.tabular_data (for "columns" tables) or
        schema.rules.sidecars (for "metadata" files).
        Only one item may be provided for columns tables.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    if isinstance(table_name, str):
        table_name = [table_name]

    elements: dict[str, str | dict[str, str]] = {}
    for table in table_name:
        if table_type == "metadata":
            table_schema = schema.rules[table]
            new_elements = table_schema.fields
        elif table_type == "columns":
            table_schema = schema.rules.tabular_data[table]
            new_elements = table_schema.columns
            # Since only one table may be provided for columns, we can just use this directly.
            additional_columns = table_schema.get("additional_columns", "not_allowed")
            initial_columns = table_schema.get("initial_columns", [])
            index_columns = table_schema.get("index_columns", [])
        else:
            raise ValueError(f"Unsupported 'table_type': '{table_type}'")

        overlap = set(new_elements) & set(elements)
        if overlap:
            raise BIDSSchemaError(
                f"Schema tables {table_name} share overlapping fields: {overlap}"
            )
        elements.update(new_elements)

    subschema = schema.objects[table_type]
    retained_elements = [f for f in elements if f in subschema]
    dropped_elements = [f for f in elements if f not in subschema]
    if dropped_elements:
        print("Warning: Missing elements: {}".format(", ".join(dropped_elements)))

    subschema = {k: v for k, v in subschema.items() if k in retained_elements}
    element_info = {}
    for element, val in elements.items():
        if isinstance(val, str):
            level = val
            level_addendum = ""
            description_addendum = ""
        else:
            level = val["level"]
            level_addendum = val.get("level_addendum", "")
            description_addendum = val.get("description_addendum", "")

        if level_addendum:
            if level_addendum.startswith(("required", "recommended", "optional")):
                level = f"{level}, but {level_addendum}"
            else:
                # Typically begins with "if"
                level = f"{level} {level_addendum}"

        if table_type == "columns" and element in index_columns:
            if len(index_columns) == 1:
                msg = f"Values in `{schema.objects.columns[element].name}`"
            else:
                cols = [f"`{schema.objects.columns[col].name}`" for col in index_columns]
                msg = f"The combination of {', '.join(cols[:-1])} and {cols[-1]}"
            description_addendum += f"\n\n\n\n{msg} MUST be unique."

        if table_type == "columns" and initial_columns:
            if element in initial_columns:
                order = initial_columns.index(element) + 1
                order_str = utils.num2words(order, to="ordinal")
                description_addendum += (
                    f"\n\n\n\nThis column must appear **{order_str}** in the file."
                )
            else:
                description_addendum += "\n\n\n\nThis column may appear anywhere in the file."

        element_info[element] = {"table_info": (level, description_addendum)}

    if table_type == "columns":
        # Add info about additional columns to table.
        # its_complicated won't add any info about additional columns.
        if additional_columns == "not_allowed":
            element_info["Additional Columns"] = {
                "table_info": ("NOT ALLOWED", "Additional columns are not allowed.")
            }
        elif additional_columns == "allowed_if_defined":
            element_info["Additional Columns"] = {
                "table_info": (
                    "OPTIONAL",
                    (
                        "Additional columns are allowed if they are defined in the associated "
                        "metadata file."
                    ),
                )
            }
        elif additional_columns == "allowed":
            element_info["Additional Columns"] = {
                "table_info": ("OPTIONAL", "Additional columns are allowed.")
            }

        subschema.update(
            {
                "Additional Columns": {
                    "name": "**Additional Columns**",
                    "description": "",
                    "type": "`n/a`",
                }
            }
        )

    table_str = _make_object_table(
        subschema,
        field_info=element_info,
        table_type=table_type,
        src_path=src_path,
        tablefmt=tablefmt,
    )
    return table_str


def make_entity_table(schema, tablefmt="github", src_path=None, **kwargs):
    """Produce entity table (markdown) based on schema.

    Parameters
    ----------
    schema : str
        Directory containing schema, which is stored in yaml files.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.

    Returns
    -------
    table_str : str
        Markdown string containing the table.
    """
    schema = Namespace(filter_schema(schema.to_dict(), **kwargs))

    # prepare the table based on the schema
    header = ["Entity"]
    formats = ["Format"]
    table = [formats]

    # Compose header and formats first
    for long_name in schema.rules.entities:
        entity = schema.objects.entities[long_name]
        header.append(entity.display_name)
        formats.append(f"[`{entity.name}-<{entity.format}>`]({ENTITIES_PATH}.md#{entity.name})")

    rows = {}
    for rule in schema.rules.files.raw.values(level=2):
        suffixes = rule.get("suffixes")
        if not suffixes:
            continue

        entities = []
        for ent in schema.rules.entities:
            val = rule.entities.get(ent, "")
            entities.append((ent, val if isinstance(val, str) else val["level"]))

        for dtype in rule.datatypes:
            row = rows.setdefault((dtype, tuple(entities)), [])
            row.extend(suffixes)

    for (dtype, entities), suffixes in rows.items():
        suf_str = " ".join(
            f"[{suffix}]({GLOSSARY_PATH}.md#objects.suffixes.{suffix})" for suffix in suffixes
        )
        # TODO: <br> is specific for html form
        fmt_str = f"[{dtype}]({GLOSSARY_PATH}.md#objects.datatypes.{dtype})<br>({suf_str})"
        table.append([fmt_str] + [level.upper() for ent, level in entities])

    # Create multi-level index because first two rows are headers
    cols = list(zip(header, table[0]))
    cols = pd.MultiIndex.from_tuples(cols)
    table = pd.DataFrame(data=table[1:], columns=cols)
    table = table.set_index(("Entity", "Format"))

    # Remove unnecessary columns
    table = utils.drop_unused_entities(table)
    table = utils.flatten_multiindexed_columns(table)

    table[table.index.name] = table.index
    table = table.set_index(table.index.name, drop=True).sort_index()

    # Print it as markdown
    table_str = tabulate(table, headers="keys", tablefmt=tablefmt)
    table_str = table_str.replace("SPEC_ROOT", utils.get_relpath(src_path))
    return table_str


def make_suffix_table(schema, suffixes, src_path=None, tablefmt="github"):
    """Produce suffix table (markdown) based on requested suffixes.

    Parameters
    ----------
    schema : dict
    suffixes : list of str
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : str

    Returns
    -------
    table_str : str
        Tabulated table as a string.
    """
    table_type = "suffixes"

    # The filter function doesn't work here.
    subschema = schema["objects"][table_type]

    all_suffixes = pd.DataFrame.from_dict(subschema, orient="index")
    df = all_suffixes[all_suffixes.value.isin(suffixes)]
    df = df.reset_index(drop=False)

    suffixes_not_found = set(suffixes) - set(df.value)
    if suffixes_not_found:
        raise Exception("Warning: Missing suffixes: {}".format(", ".join(suffixes_not_found)))

    def preproc_desc(desc):
        return (
            desc.replace("\\\n", "")  # A backslash before a newline means continue a string
            .replace("\n\n", "<br>")  # Two newlines should be respected
            .replace("\n", " ")  # Otherwise a newline corresponds to a space
        )

    def preproc_suffix(row):
        return f"[{row['value']}]({GLOSSARY_PATH}.md#objects.{table_type}.{row['index']})"

    df.description = df.description.apply(preproc_desc)
    df["suffix"] = df.apply(preproc_suffix, axis=1)
    df = df[["suffix", "display_name", "description"]]
    df.columns = ["`suffix`", "**Name**", "**Description**"]
    df = df.reset_index(drop=False)
    df = df.set_index("**Name**")
    df = df[["`suffix`", "**Description**"]]

    # Print it as markdown
    table_str = tabulate(df, headers="keys", tablefmt=tablefmt)
    # Spec internal links need to be replaced
    table_str = table_str.replace("SPEC_ROOT", utils.get_relpath(src_path))
    return table_str


def make_json_table(
    schema: Namespace,
    table_name: str | list[str],
    src_path: str | None = None,
    tablefmt: str = "github",
):
    """Produce metadata table (markdown) based on requested fields.

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    table_name : str or list of str
        Qualified name(s) in schema.rules.sidecars
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    table_str = _make_table_from_rule(
        schema=schema,
        table_type="metadata",
        table_name=table_name,
        src_path=src_path,
        tablefmt=tablefmt,
    )

    return table_str


def make_sidecar_table(
    schema: Namespace,
    table_name: str | list[str],
    src_path: str | None = None,
    tablefmt: str = "github",
):
    """Produce metadata table (markdown) based on requested fields.

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    table_name : str or list of str
        Qualified name(s) in schema.rules.sidecars
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    table_str = _make_table_from_rule(
        schema=schema,
        table_type="metadata",
        table_name=[
            f"sidecars.{table}"
            for table in ([table_name] if isinstance(table_name, str) else table_name)
        ],
        src_path=src_path,
        tablefmt=tablefmt,
    )

    return table_str


def make_metadata_table(schema, field_info, src_path=None, tablefmt="github"):
    """Produce metadata table (markdown) based on requested fields.

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    field_info : dict of strings or tuples
        A dictionary mapping metadata keys to requirement levels in the
        rendered metadata table.
        The dictionary values may be strings, in which case the string
        is the requirement level information, or two-item tuples of strings,
        in which case the first string is the requirement level information
        and the second string is additional table-specific information
        about the metadata field that will be appended to the field's base
        definition from the schema.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    fields = list(field_info.keys())
    # The filter function doesn't work here.
    table_type = "metadata"
    subschema = schema["objects"][table_type]

    retained_fields = [f for f in fields if f in subschema.keys()]
    dropped_fields = [f for f in fields if f not in subschema.keys()]
    if dropped_fields:
        print("Warning: Missing fields: {}".format(", ".join(dropped_fields)))

    subschema = {k: v for k, v in subschema.items() if k in retained_fields}
    # inverted_schema = {v["name"]: k for k, v in subschema.items()}

    # Reformat the field info (requirement level and description addendum)
    field_info_2 = {}
    for field, val in field_info.items():
        if isinstance(val, tuple):
            field_info_2[field] = {"table_info": val}
        else:
            field_info_2[field] = {"table_info": (val, "")}

    table_str = _make_object_table(
        subschema,
        field_info=field_info_2,
        table_type=table_type,
        src_path=src_path,
        tablefmt=tablefmt,
    )

    return table_str


def make_subobject_table(
    schema: Namespace,
    object_name: str,
    src_path: str | None = None,
    tablefmt: str = "github",
):
    """Create a metadata table (markdown) based on the properties of an object

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    object_name : str
        Qualified name in schema.objects
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    obj = schema.objects[object_name]
    required_fields = set(obj.get("required", ()))
    recommended_fields = set(obj.get("recommended", ()))

    field_info = {}
    for field in obj.properties:
        if field in required_fields:
            req_level = "required"
        elif field in recommended_fields:
            req_level = "recommended"
        else:
            req_level = "optional"
        field_info[field] = {"table_info": (req_level, "")}

    table_str = _make_object_table(
        obj.properties,
        field_info=field_info,
        table_type="subobject",
        src_path=src_path,
        tablefmt=tablefmt,
        n_values_to_combine=10000,  # no combination
    )

    return table_str


def make_columns_table(
    schema: Namespace,
    table_name: str,
    src_path: str | None = None,
    tablefmt: str = "github",
):
    """Produce columns table (markdown) based on requested fields.

    Parameters
    ----------
    schema : Namespace
        The BIDS schema.
    table_name : str
        Qualified name in schema.rules.tabular_data.
        Only one table may be provided in this function.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    tablefmt : string, optional
        The target table format. The default is "github" (GitHub format).

    Returns
    -------
    table_str : str
        The tabulated table as a Markdown string.
    """
    table_str = _make_table_from_rule(
        schema=schema,
        table_type="columns",
        table_name=table_name,
        src_path=src_path,
        tablefmt=tablefmt,
    )

    return table_str



"""Namespace types

The purpose of the :class:`~Namespace` type is to make a directory of
YAML files available as a single dictionary and allow attribute (``.``)
lookups.
"""

import json
import typing as ty
from collections.abc import ItemsView, KeysView, Mapping, MutableMapping, ValuesView
from pathlib import Path


def _expand_dots(entry: tuple[str, ty.Any]) -> tuple[str, ty.Any]:
    # Helper function for expand
    key, val = entry
    if "." in key:
        init, post = key.split(".", 1)
        return init, dict([_expand_dots((post, val))])
    return key, expand(val)


def expand(element):
    """Expand a dict, recursively, to replace dots in keys with recursive dictionaries

    Parameters
    ----------
    element : dict
        The dictionary to expand

    Returns
    -------
    dict :
        The expanded dictionary

    Examples
    --------
    >>> expand({"a": 1, "b.c": 2, "d": [{"e": 3, "f.g": 4}]})
    {'a': 1, 'b': {'c': 2}, 'd': [{'e': 3, 'f': {'g': 4}}]}
    """
    if isinstance(element, dict):
        return {key: val for key, val in map(_expand_dots, element.items())}
    elif isinstance(element, list):
        return [expand(el) for el in element]
    return element


class NsItemsView(ItemsView):
    def __init__(self, namespace, level):
        self._mapping = namespace
        self._level = level

    def __contains__(self, item):
        key, val = item
        keys = key.split(".", self._level - 1)
        if "." in keys[-1]:
            return False
        return self._mapping[key] == val

    def __iter__(self):
        l1 = ItemsView(self._mapping)
        if self._level == 1:
            yield from l1
        else:
            yield from (
                (f"{key}.{subkey}", subval)
                for key, val in l1
                # Items/keys/values can only be found in namespaces
                # ignore lists and scalars
                if isinstance(val, Mapping)
                for subkey, subval in NsItemsView(val, self._level - 1)
            )


class NsKeysView(KeysView):
    def __init__(self, namespace, level):
        self._mapping = namespace
        self._level = level

    def __contains__(self, key):
        keys = key.split(".", self._level - 1)
        if "." in keys[-1]:
            return False
        return key in self._mapping

    def __iter__(self):
        yield from (key for key, val in NsItemsView(self._mapping, self._level))


class NsValuesView(ValuesView):
    def __init__(self, namespace, level):
        self._mapping = namespace
        self._level = level
        self._items = NsItemsView(namespace, level)

    def __contains__(self, val):
        return any(val == item[1] for item in self._items)

    def __iter__(self):
        yield from (val for key, val in self._items)


class Namespace(MutableMapping):
    """Provides recursive attribute style access to a dict-like structure

    Examples
    --------
    >>> ns = Namespace.build({"a": 1, "b.c": "val"})
    >>> ns.a
    1
    >>> ns["a"]
    1
    >>> ns.b
    <Namespace {'c': 'val'}>
    >>> ns["b"]
    <Namespace {'c': 'val'}>
    >>> ns.b.c
    'val'
    >>> ns["b.c"]
    'val'
    >>> ns["b"]["c"]
    'val'
    >>> ns.b["c"]
    'val'
    >>> ns["b"].c
    'val'

    ``.keys()``, ``.values()`` and ``.items()`` can take an optional ``level`` argument:

    >>> list(ns.keys())
    ['a', 'b']
    >>> list(ns.keys(level=2))
    ['b.c']
    >>> 'b.c' in ns.keys()
    False
    >>> 'b.c' in ns.keys(level=2)
    True

    >>> list(ns.values())
    [1, <Namespace {'c': 'val'}>]
    >>> list(ns.values(level=2))
    ['val']
    >>> 'val' in ns.values()
    False
    >>> 'val' in ns.values(level=2)
    True

    >>> list(ns.items())
    [('a', 1), ('b', <Namespace {'c': 'val'}>)]
    >>> list(ns.items(level=2))
    [('b.c', 'val')]
    >>> ("b.c", "val") in ns.items()
    False
    >>> ("b.c", "val") in ns.items(level=2)
    True

    >>> ns["d.e.f"] = "val2"
    >>> ns.d.e
    <Namespace {'f': 'val2'}>
    >>> ns.d.e.f
    'val2'
    >>> del ns['d']
    """

    def __init__(self, *args, **kwargs):
        self._properties = dict(*args, **kwargs)

    def to_dict(self) -> dict:
        def _to_dict(obj):
            if isinstance(obj, Namespace):
                return {k: _to_dict(v) for k, v in obj._properties.items()}
            if isinstance(obj, list):
                return [_to_dict(v) for v in obj]
            if isinstance(obj, dict):
                return {k: _to_dict(v) for k, v in obj.items()}
            return obj

        return _to_dict(self)

    def __deepcopy__(self, memo):
        return self.build(self.to_dict())

    @classmethod
    def view(cls, mapping):
        if isinstance(mapping, cls):
            return mapping
        if not isinstance(mapping, dict):
            raise ValueError("Namespace.view can only be made from a dict")
        new = cls()
        new._properties = mapping
        return new

    @classmethod
    def build(cls, mapping):
        """Expand mapping recursively and return as namespace"""
        return cls(expand(mapping))

    def items(self, *, level=1) -> NsItemsView:
        return NsItemsView(self, level)

    def keys(self, *, level=1) -> NsKeysView:
        return NsKeysView(self, level)

    def values(self, *, level=1) -> NsValuesView:
        return NsValuesView(self, level)

    def __getattribute__(self, key):
        # Return actual properties first
        err = None
        try:
            return super().__getattribute__(key)
        except AttributeError as e:
            err = e

        # Utilize __getitem__ but keep original error on failure
        try:
            return self[key]
        except KeyError:
            raise err

    def _get_mapping(self, key: str) -> tuple[Mapping, str]:
        subkeys = key.split(".")
        mapping = self._properties
        for subkey in subkeys[:-1]:
            mapping = mapping.setdefault(subkey, {})
            mapping = getattr(mapping, "_properties", mapping)
            if not isinstance(mapping, Mapping):
                raise KeyError(f"{key} (subkey={subkey})")
        return mapping, subkeys[-1]

    def __getitem__(self, key):
        mapping, subkey = self._get_mapping(key)
        val = mapping[subkey]
        if isinstance(val, dict):
            val = self.view(val)
        return val

    def __setitem__(self, key, val):
        mapping, subkey = self._get_mapping(key)
        mapping[subkey] = val

    def __delitem__(self, key):
        mapping, subkey = self._get_mapping(key)
        del mapping[subkey]

    def __repr__(self):
        return f"<Namespace {self._properties}>"

    def __len__(self):
        return len(self._properties)

    def __iter__(self):
        return iter(self._properties)

    @classmethod
    def from_directory(cls, path, fmt="yaml"):
        if fmt == "yaml":
            if isinstance(path, str):
                path = Path(path)
            return cls.build(_read_yaml_dir(path))
        raise NotImplementedError(f"Unknown format: {fmt}")

    def to_json(self, **kwargs) -> str:
        return json.dumps(self, cls=MappingEncoder, **kwargs)

    @classmethod
    def from_json(cls, jsonstr: str):
        return cls.build(json.loads(jsonstr))


def _read_yaml_dir(path: Path) -> dict:
    mapping = {}
    for subpath in sorted(path.iterdir()):
        if subpath.is_dir():
            mapping[subpath.name] = _read_yaml_dir(subpath)
        elif subpath.name.endswith("yaml"):
            import yaml

            mapping[subpath.stem] = yaml.safe_load(subpath.read_text())
    return mapping


class MappingEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            return super().default(o)
        except TypeError as e:
            err = e
        if isinstance(o, Mapping):
            return dict(o)
        raise err



"""Tests for the bidsschematools package."""

import os

import pytest

from bidsschematools.render import text


def test_make_entity_definitions(schema_obj):
    """
    Test whether expected format strings are present.
    This should be stable with respect to schema format.
    """
    schema_text = text.make_entity_definitions(schema_obj)
    expected_formats = [
        "**Format**: `sub-<label>`",
        "**Format**: `ses-<label>`",
        "**Format**: `sample-<label>`",
        "**Format**: `task-<label>`",
        "**Format**: `acq-<label>`",
        "**Format**: `ce-<label>`",
        "**Format**: `trc-<label>`",
        "**Format**: `stain-<label>`",
        "**Format**: `rec-<label>`",
        "**Format**: `dir-<label>`",
        "**Format**: `run-<index>`",
        "**Format**: `mod-<label>`",
        "**Format**: `echo-<index>`",
        "**Format**: `flip-<index>`",
        "**Format**: `inv-<index>`",
        "**Format**: `mt-<label>`",
    ]
    for expected_format in expected_formats:
        assert expected_format in schema_text


def test_make_glossary(schema_obj, schema_dir):
    """
    Test whether files under the schema objects subdirectory correspond to entries, and
    that rules are not mis-loaded as objects.
    This may need to be updated for schema hierarchy changes.
    """
    # Test consistency
    object_files = []
    for root, dirs, files in os.walk(schema_dir, topdown=False):
        if "objects" in root:
            for object_file in files:
                object_base, _ = os.path.splitext(object_file)
                object_files.append(object_base)
    rule_files = []
    for root, dirs, files in os.walk(schema_dir, topdown=False):
        if "rules" in root:
            for rule_file in files:
                rule_base, _ = os.path.splitext(rule_file)
                rule_files.append(rule_base)
    rules_only = list(filter(lambda a: a not in object_files, rule_files))

    glossary = text.make_glossary(schema_obj)
    for line in glossary.split("\n"):
        if line.startswith('<a name="objects.'):
            # Are all objects objects?
            assert any([line.startswith(f'<a name="objects.{i}') for i in object_files])
            # Are rules loaded incorrectly?
            assert not any([line.startswith(f'<a name="objects.{i}') for i in rules_only])


def test_make_filename_template(schema_obj, schema_dir):
    """
    Test whether:

    * The base hierarchy structure of mandatory subject and optional session is
      returned. This should be robust with respect to schema format.
    * Each directory contains at least one possible pattern.
      This should be robust with respect to schema format.
    * All files under the datatype rules subdirectory have corresponding entries.
      This may need to be updated for schema hierarchy changes.
    """
    filename_template = text.make_filename_template("raw", schema_obj, pdf_format=True)

    # Test predefined substrings
    expected_template_part = """
sub-<label>/
    [ses-<label>/]
        anat/
    """
    assert expected_template_part in filename_template

    # Test consistency
    datatypes = {
        datatype
        for rule in schema_obj.rules.files.raw.values(level=2)
        for datatype in rule.datatypes
    }

    datatype_count = len(datatypes)
    datatype_bases = [f"        {i}/" for i in datatypes]
    datatype_level = False
    datatype_file_start = "            sub-<label>"
    datatype_bases_found = 0
    for line in filename_template.split("\n"):
        if datatype_level:
            # Is there at least one file pattern per datatype?
            assert line.startswith(datatype_file_start)
            datatype_level = False
        if line in datatype_bases:
            datatype_level = True
            datatype_bases_found += 1
    # Are all datatypes listed?
    assert datatype_bases_found == datatype_count


def test_define_common_principles(schema_obj):
    """Ensure that define_common_principles returns a string."""
    common_principles_str = text.define_common_principles(schema_obj)
    # Must be a string
    assert isinstance(common_principles_str, str)
    # Must be fairly long
    assert len(common_principles_str) > 100


def test_append_filename_template_legend():
    """Check contents of generated string."""
    test_str = text.append_filename_template_legend("", pdf_format=False)
    assert isinstance(test_str, str)
    assert "follow the links" in test_str

    test_str = text.append_filename_template_legend("", pdf_format=True)
    assert isinstance(test_str, str)
    assert "follow the links" not in test_str


def test_define_allowed_top_directories(schema_obj):
    """smoke test for allowed top directories."""
    test_str = text.define_allowed_top_directories(schema_obj)
    assert isinstance(test_str, str)


def test_render_text(schema_obj):
    test_str = text.render_text(
        schema_obj, key="objects.files.dataset_description.description", src_path=None
    )
    assert (
        test_str == "The file `dataset_description.json` is a JSON file describing the dataset.\n"
    )


def test_render_text_errors(schema_obj):
    with pytest.raises(ValueError, match="does not refer to a text field"):
        text.render_text(schema_obj, key="dataset_description", src_path=None)



"""Schema loading- and processing-related functions."""

import json
import os
import re
from collections.abc import Iterable, Mapping
from copy import deepcopy
from functools import lru_cache
from pathlib import Path

from . import data, utils
from .types import Namespace

lgr = utils.get_logger()


class BIDSSchemaError(Exception):
    """Errors indicating invalid values in the schema itself"""


def _get_schema_version(schema_dir):
    """
    Determine schema version for given schema directory, based on file specification.
    """

    schema_version_path = os.path.join(schema_dir, "SCHEMA_VERSION")
    with open(schema_version_path) as f:
        schema_version = f.readline().rstrip()
    return schema_version


def _get_bids_version(schema_dir):
    """
    Determine BIDS version for given schema directory, with directory name, file specification,
    and string fallback.
    """

    bids_version_path = os.path.join(schema_dir, "BIDS_VERSION")
    try:
        with open(bids_version_path) as f:
            bids_version = f.readline().rstrip()
    # If this file is not in the schema, fall back to placeholder heuristics:
    except FileNotFoundError:
        # Maybe the directory encodes the version, as in:
        # https://github.com/bids-standard/bids-schema
        _, bids_version = os.path.split(schema_dir)
        if not re.match(r"^.*?[0-9]*?\.[0-9]*?\.[0-9]*?.*?$", bids_version):
            # Then we don't know, really.
            bids_version = schema_dir
    return bids_version


def _find(obj, predicate):
    """Find objects in an arbitrary object that satisfy a predicate.

    Note that this does not cut branches, so every iterable sub-object
    will be fully searched.

    Parameters
    ----------
    obj : object
    predicate : function

    Returns
    -------
    generator
        A generator of entries in ``obj`` that satisfy the predicate.
    """
    try:
        if predicate(obj):
            yield obj
    except Exception:
        pass

    iterable = ()
    if isinstance(obj, Mapping):
        iterable = obj.values()
    elif not isinstance(obj, str) and isinstance(obj, Iterable):
        iterable = obj

    for item in iterable:
        yield from _find(item, predicate)


def _dereference(namespace, base_schema):
    # In-place, recursively dereference objects
    # This allows a referenced object to itself contain a reference
    # A dependency graph could be constructed, but would likely be slower
    # to build than to duplicate a couple dereferences
    for struct in _find(namespace, lambda obj: "$ref" in obj):
        target = base_schema.get(struct["$ref"])
        if target is None:
            raise ValueError(f"Reference {struct['$ref']} not found in schema.")
        if isinstance(target, Mapping):
            struct.pop("$ref")
            _dereference(target, base_schema)
            struct.update({**target, **struct})


def dereference(namespace, inplace=True):
    """Replace references in namespace with the contents of the referred object.

    Parameters
    ----------
    namespace : Namespace
        Namespace for which to dereference.

    inplace : bool, optional
        Whether to modify the namespace in place or create a copy, by default True.

    Returns
    -------
    namespace : Namespace
        Dereferenced namespace
    """
    if not inplace:
        namespace = deepcopy(namespace)

    _dereference(namespace, namespace)

    # At this point, any remaining refs are one-off objects in lists
    for struct in _find(namespace, lambda obj: any("$ref" in sub for sub in obj)):
        for i, item in enumerate(struct):
            try:
                target = item.pop("$ref")
            except (AttributeError, KeyError):
                pass
            else:
                struct[i] = namespace.get(target)

    return namespace


def flatten_enums(namespace, inplace=True):
    """Replace enum collections with a single enum, merging enums contents.

    The function helps reducing the complexity of the schema by assuming
    that the values in the conditions (anyOf) are mutually exclusive.

    Parameters
    ----------
    namespace : dict
        Schema in dictionary form to be flattened.

    Returns
    -------
    schema : dict
        Schema with flattened enums.

    Examples
    --------

    >>> struct = {
    ...   "anyOf": [
    ...      {"type": "string", "enum": ["A", "B", "C"]},
    ...      {"type": "string", "enum": ["D", "E", "F"]},
    ...   ]
    ... }
    >>> flatten_enums(struct)
    {'type': 'string', 'enum': ['A', 'B', 'C', 'D', 'E', 'F']}
    """
    if not inplace:
        namespace = deepcopy(namespace)
    for struct in _find(namespace, lambda obj: "anyOf" in obj):
        try:
            # Deduplicate because JSON schema validators may not like duplicates
            # Long run, we should get rid of this function and have the rendering
            # code handle anyOfs
            all_enum = list(dict.fromkeys(val for item in struct["anyOf"] for val in item["enum"]))
        except KeyError:
            continue

        del struct["anyOf"]
        struct.update({"type": "string", "enum": all_enum})
    return namespace


@lru_cache
def load_schema(schema_path=None):
    """Load the schema into a dictionary.

    This function allows the schema, like BIDS itself, to be specified in
    a hierarchy of directories and files.
    Filenames (minus extensions) and directory names become keys
    in the associative array (dict) of entries composed from content
    of files and entire directories.

    Parameters
    ----------
    schema_path : str, optional
        Directory containing yaml files or yaml file. If ``None``, use the
        default schema packaged with ``bidsschematools``.

    Returns
    -------
    dict
        Schema in dictionary form.

    Notes
    -----
    This function is cached, so it will only be called once per schema path.
    """
    if schema_path is None:
        # Default to bundled JSON, fall back to bundled YAML directory
        schema_path = data.load.readable("schema.json")
        if not schema_path.is_file():
            schema_path = data.load.readable("schema")

            # Probably a Windows checkout with a git link. Resolve first.
            if schema_path.is_file() and (content := schema_path.read_text()).startswith("../"):
                schema_path = Path.resolve(schema_path.parent / content)
        lgr.info("No schema path specified, defaulting to the bundled schema, `%s`.", schema_path)
    elif isinstance(schema_path, str):
        schema_path = Path(schema_path)

    # JSON file: just load it
    if schema_path.is_file():
        return Namespace.from_json(schema_path.read_text())

    # YAML directory: load, dereference and set versions
    schema = Namespace.from_directory(schema_path)
    if not schema.objects:
        raise ValueError(f"objects subdirectory path not found in {schema_path}")
    if not schema.rules:
        raise ValueError(f"rules subdirectory path not found in {schema_path}")

    dereference(schema)
    flatten_enums(schema)

    schema["bids_version"] = _get_bids_version(schema_path)
    schema["schema_version"] = _get_schema_version(schema_path)

    return schema


def export_schema(schema):
    """Export the schema to JSON format.

    Parameters
    ----------
    schema : dict
        The schema object, in dictionary form.

    Returns
    -------
    json : str
        The schema serialized as a JSON string.
    """
    return schema.to_json()


def filter_schema(schema, **kwargs):
    """Filter the schema based on a set of keyword arguments.

    Parameters
    ----------
    schema : dict
        The schema object, which is a dictionary with nested dictionaries and
        lists stored within it.

    Other Parameters
    ----------------
    **kwargs : dict
        Keyword arguments used to filter the schema.
        Example kwargs that may be used include: "suffixes", "datatypes",
        "extensions".

    Returns
    -------
    new_schema : dict
        The filtered version of the schema.

    Notes
    -----
    This function calls itself recursively, in order to apply filters at
    arbitrary depth.

    Warning
    -------
    This function employs a *very* simple filter. It is very limited.
    """
    new_schema = deepcopy(schema)
    if isinstance(new_schema, dict):
        # Reduce values in dict to only requested
        for k, v in kwargs.items():
            if k in new_schema.keys():
                filtered_item = deepcopy(new_schema[k])
                if isinstance(filtered_item, dict):
                    filtered_item = {k1: v1 for k1, v1 in filtered_item.items() if k1 in v}
                else:
                    filtered_item = [i for i in filtered_item if i in v]
                new_schema[k] = filtered_item

            for k2, v2 in new_schema.items():
                new_schema[k2] = filter_schema(new_schema[k2], **kwargs)

    elif isinstance(new_schema, list):
        for i, item in enumerate(new_schema):
            if isinstance(item, dict):
                new_schema[i] = filter_schema(item, **kwargs)
    return new_schema


def validate_schema(schema: Namespace):
    """Validate a schema against the BIDS metaschema."""
    try:
        from jsonschema import ValidationError, validate
    except ImportError as e:
        raise RuntimeError(
            "The `jsonschema` package is required to validate schemas. "
            "Please install it with `pip install jsonschema`."
        ) from e

    from .data import load

    metaschema = json.loads(load.readable("metaschema.json").read_text())

    # validate is put in this try/except clause because the error is sometimes too long to
    # print in the terminal
    try:
        validate(instance=schema.to_dict(), schema=metaschema)
    except ValidationError as e:
        import tempfile

        with tempfile.NamedTemporaryFile(
            prefix="schema_error_", suffix=".txt", delete=False, mode="w+"
        ) as file:
            file.write(str(e))
            # ValidationError does not have an add_note method yet
            # e.add_note(f"See {file.name} for full error log.")
        raise e



"""Simple validation tests on schema rules."""

import warnings
from collections.abc import Mapping

import pytest


def _dict_key_lookup(_dict, key, path=[]):
    """Look up any uses of a key in a nested dictionary.

    Adapted from https://stackoverflow.com/a/60377584/2589328.
    """
    results = []
    if isinstance(_dict, Mapping):
        if key in _dict:
            results.append((path + [key], _dict[key]))

        for k, v in _dict.items():
            results.extend(_dict_key_lookup(v, key, path=path + [k]))

    elif isinstance(_dict, list):
        for index, item in enumerate(_dict):
            results.extend(_dict_key_lookup(item, key, path=path + [index]))

    return results


@pytest.mark.validate_schema
def test_rule_objects(schema_obj):
    """Ensure that all objects referenced in the schema rules are defined in
    their object portion.

    In the case of datatypes, the key and "value" field are always the same.

    Some other object types, such as associated_data, common_principles, formats, modalities,
    and top_level_files, are not checked in the rules at all.

    Additionally, this test only checks rules that fit the keys.
    """
    OBJECT_TYPE_MAPPER = {
        "metadata": "fields",  # metadata in objects is referred to as fields in rules
    }

    not_found = []  # A list of undefined, but referenced, objects
    for object_type in schema_obj.objects:
        # "files" is both an object name and a grouping of rules
        # The next line would be a false positive hit
        if object_type == "files":
            continue
        # Find all uses of a given object type in the schema rules
        type_instances_in_rules = _dict_key_lookup(
            schema_obj.rules,
            OBJECT_TYPE_MAPPER.get(object_type, object_type),
        )
        if not type_instances_in_rules:
            continue

        for type_instance in type_instances_in_rules:
            path, instance = type_instance
            is_list = True
            if isinstance(instance, Mapping):
                instance = list(instance)
                is_list = False

            for i_use, use in enumerate(instance):
                if use == "derivatives":
                    # Skip derivatives dirs, because the dir is treated as a "use" instead.
                    continue
                elif "[]" in use:
                    # Rules may reference metadata fields with lists.
                    # This test can't handle this yet, so skip.
                    continue
                elif "{}" in use:
                    # Rules may reference sub-dictionaries in metadata fields.
                    # This test can't handle this yet, so skip.
                    continue

                if object_type in ["extensions", "suffixes"]:
                    # Some object types are referenced via their "value" fields in the rules
                    object_values = [
                        value["value"] for value in schema_obj["objects"][object_type].values()
                    ]
                else:
                    # But other object types are referenced via their keys
                    object_values = list(schema_obj["objects"][object_type].keys())

                # Build a list of items mentioned in rules, but not found in objects.
                if use not in object_values:
                    if (use, object_type) == ("phenotype", "datatypes"):
                        # Special case: phenotype is a top-level directory
                        # that acts like a datatype, but we don't want to
                        # define it that way in the glossary, currently.
                        continue
                    temp_path = path[:]
                    if is_list:
                        temp_path[-1] += f"[{i_use}]"

                    not_found.append((temp_path, use))

    if not_found:
        not_found_string = "\n".join([f"{'.'.join(path)} == {val}" for path, val in not_found])
        raise AssertionError(f"Undefined objects found in rules: {not_found_string}")


@pytest.mark.validate_schema
def test_entity_order(schema_obj):
    """Check the order of the entities of the suffix group of each datatype
    and lists those that are out of order.
    """
    status_ok = True

    entities_order = schema_obj.rules.entities

    for key, group in schema_obj.rules.files.items(level=2):
        print(f"Checking {key}")
        entities = list(group.get("entities", ()))
        correct_order = sorted(entities, key=lambda x: entities_order.index(x))

        if entities != correct_order:
            status_ok = False
            warnings.warn(
                f"""\n\nfilename rule {key} has entities out-of-order:
                - got: {entities}
                - should be: {correct_order}
                """
            )

    if not status_ok:
        raise RuntimeError(
            "Some suffix groups have their entities out of order. See warnings above."
        )



"""Test data module

.. autofunction:: load_test_data

.. automethod:: load_test_data.readable

.. automethod:: load_test_data.as_path

.. automethod:: load_test_data.cached
"""

from acres import Loader

__all__ = ("load_test_data",)

load_test_data = Loader(__spec__.name)



"""Utility functions for specification rendering tools."""

import math
import posixpath


def _link_with_html(string, html_path=None, heading=None, pdf_format=False):
    """Wrap a string in an HTML hyperlink.

    Parameters
    ----------
    string : str
        The string to wrap a hyperlink around.
    html_path : None or str, optional
        Path to the HTML file that the string should link to.
    heading : None or str, optional
        The heading on the HTML page the string should link to.
    pdf_format : bool, optional
        If True, the string will be returned unmodified.
        If False, a hyperlink will be generated around the string,
        linking to the ``heading`` heading in the ``html_path`` page.
        Default is False.

    Returns
    -------
    string : str
        The modified (or unmodified) string.
    """
    if not pdf_format:
        string = string.replace("<", "&lt;").replace(">", "&gt;")
        string = f'<a href="{html_path}#{heading}">{string}</a>'

    return string


def combine_extensions(lst, html_path=None, heading_lst=None, pdf_format=True):
    """Combine extensions with their compressed versions in a list.

    Valid combinations are hardcoded in the function,
    since some extensions look like compressed versions of one another, but are not.

    Parameters
    ----------
    lst : list of str
        Raw list of extensions.
    html_path : None or str
        Path to the HTML file that each extension should link to.
        Only used if pdf_format is False.
        Default is None.
    heading_lst : None or list of str
        List of headings in the HTML page to link to.
        Should be one heading for each extension in lst.
        Only used if pdf_format is False.
        Default is None.
    pdf_format : bool, optional
        If True, the extensions will be compiled as markdown strings,
        without any hyperlinks, so that the specification's PDF build will look right.
        If False, the extensions will use HTML and include hyperlinks to the their
        associated glossary entries.
        This works on the website.
        Default is True.

    Returns
    -------
    new_lst : list of str
        List of extensions, with compressed and uncompressed versions of the same extension
        combined.
    """
    COMPRESSION_EXTENSIONS = [".gz"]
    if pdf_format and not heading_lst:
        heading_lst = lst[:]

    new_lst = []
    items_to_remove = []
    for i_item, item in enumerate(lst):
        for ext in COMPRESSION_EXTENSIONS:
            if item.endswith(ext) and item.replace(ext, "") in lst:
                base_item_idx = lst.index(item.replace(ext, ""))
                temp_item = _link_with_html(
                    lst[base_item_idx],
                    html_path=html_path,
                    heading=heading_lst[base_item_idx].lower(),
                    pdf_format=pdf_format,
                )
                ext_string = _link_with_html(
                    ext,
                    html_path=html_path,
                    heading=heading_lst[i_item].lower(),
                    pdf_format=pdf_format,
                )

                temp_item = temp_item + "[" + ext_string + "]"
                new_lst.append(temp_item)
                items_to_remove.append(item)
                items_to_remove.append(item.replace(ext, ""))

    heading_lst = [head for i, head in enumerate(heading_lst) if lst[i] not in items_to_remove]
    items_to_add = [item for item in lst if item not in items_to_remove]
    item_strings_to_add = []
    for i_item, item in enumerate(items_to_add):
        item_strings_to_add.append(
            _link_with_html(
                item,
                html_path=html_path,
                heading=heading_lst[i_item],
                pdf_format=pdf_format,
            )
        )

    new_lst += item_strings_to_add

    return new_lst


def drop_unused_entities(df):
    """Remove columns from a dataframe where all values in the column are NaNs.

    For entity tables, this limits each table to only entities that are used
    within the modality.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing entities and datatypes/suffixes.
        Rows are datatype/suffix combinations and columns are entities.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with columns associated with unused entities removed.
    """
    df = df.replace("", math.nan).dropna(axis=1, how="all").fillna("")
    return df


def flatten_multiindexed_columns(df):
    """Remove multi-indexing of multi-indexed column headers.

    The first layer is the "DataType", while the second layer is the "Format".
    This second layer will become a new row.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with two header levels: "Datatype" and "Format".

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with the second header level ("Format") converted to a
        normal row.
    """
    # Flatten multi-index
    vals = df.index.tolist()
    df.loc["Format"] = df.columns.get_level_values(1)
    df.columns = df.columns.get_level_values(0)
    df = df.loc[["Format"] + vals]
    df.index.name = "Entity"
    return df


def get_link(string):
    """Return a hyperlink to the JSON specification for a given JSON type.

    Parameters
    ----------
    string : str
        The JSON type to link to.

    Returns
    -------
    url : str
        The hyperlink to the JSON specification for the given JSON type.
    """
    refs = {
        "array": "https://www.w3schools.com/js/js_json_arrays.asp",
        "string": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "number": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "object": "https://www.json.org/json-en.html",
        "integer": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "boolean": "https://www.w3schools.com/js/js_json_datatypes.asp",
    }
    # Allow plurals (e.g., strings -> links to string)
    dtype = string[:-1] if string[-1] == "s" else string
    url = refs.get(dtype)
    if url:
        return f"[{string}]({url})"
    return string


def resolve_metadata_type(definition):
    """Generate string of metadata type from dictionary.

    Parameters
    ----------
    definition : :obj:`dict`
        A schema object definition for a metadata term.

    Returns
    -------
    string : str
        A string describing the valid value types for the metadata term.
    """
    if "type" in definition.keys():
        string = get_link(definition["type"])

        if definition.get("enum") == ["n/a"]:
            # Special string case of n/a
            string = '`"n/a"`'

        elif "type" in definition.get("items", {}):
            # Items within arrays
            string += " of " + get_link(definition["items"]["type"] + "s")

        elif "type" in definition.get("additionalProperties", {}):
            # Values within objects
            string += " of " + get_link(definition["additionalProperties"]["type"] + "s")

    elif "anyOf" in definition:
        # Use dictionary to get unique substrings while preserving insertion order
        substrings = {resolve_metadata_type(subdict): None for subdict in definition["anyOf"]}

        string = " or ".join(substrings)

    elif "definition" in definition:
        json_def = definition["definition"]

        if "Delimiter" in json_def:
            # Delimiter indicates the value must be parsed. For BIDS purposes,
            # this is a string, even if the parsed array is of numbers.
            string = "string"
        elif "Levels" in json_def:
            # JSON keys are always strings.
            string = "string"
        elif "Units" in json_def:
            # Values with units are always (any exceptions?) numbers.
            string = "number"
        else:
            string = "string or number"
    else:
        # This clause should only catch $refs.
        # The schema should be deferenced by this point, so $refs should not exist.
        raise ValueError(f"Type could not be inferred for {definition['name']}")

    return string


def describe_valid_values(definition):
    """Build a sentence describing valid values for an object from its definition.

    This only covers booleans, enums, integers, and numbers.
    Currently uncovered are anyOfs, arrays, and objects.

    Parameters
    ----------
    definition : :obj:`dict`
        An object definition, following the BIDS schema object rules.

    Returns
    -------
    str : A sentence describing valid values for the object.
    """
    description = ""
    if "anyOf" in definition:
        return description

    if "definition" in definition:
        levels = definition["definition"].get("Levels")
        if levels:
            description = (
                f"Unless redefined in a sidecar file, must be one of: {', '.join(levels)}."
            )
        return description

    if definition["type"] == "boolean":
        description = 'Must be one of: `"true"`, `"false"`.'

    elif definition["type"] == "string":
        if "enum" in definition.keys():
            # Allow enums to be "objects" (dicts) or strings
            enums = [list(v.keys())[0] if isinstance(v, dict) else v for v in definition["enum"]]
            enums = [f'`"{v}"`' for v in enums]
            description = f"Must be one of: {', '.join(enums)}."

    elif definition["type"] in ("integer", "number"):
        minstr = maxstr = minmaxstr = ""

        if "minimum" in definition.keys():
            minstr = f"greater than or equal to {definition['minimum']}"
        elif "exclusiveMinimum" in definition.keys():
            minstr = f"greater than {definition['exclusiveMinimum']}"

        if "maximum" in definition.keys():
            maxstr = f"less than or equal to {definition['maximum']}"
        elif "exclusiveMaximum" in definition.keys():
            maxstr = f"less than {definition['exclusiveMaximum']}"

        if minstr and maxstr:
            minmaxstr = f"{minstr} and {maxstr}"
        elif minstr or maxstr:
            minmaxstr = minstr + maxstr

        if minmaxstr:
            description = f"Must be a number {minmaxstr}."

    return description


def get_relpath(src_path):
    """Retrieve relative path to the source root from the perspective of a Markdown file.

    As a convenience, ``None`` is interpreted as the empty string, and a value of ``'.'``
    is returned.

    Examples
    --------
    >>> get_relpath("02-common-principles.md")
    '.'
    >>> get_relpath("04-modality-specific-files/01-magnetic-resonance-imaging-data.md")
    '..'
    >>> get_relpath("we/lack/third_levels.md")
    '../..'
    >>> get_relpath(None)
    '.'
    """
    return posixpath.relpath(".", posixpath.dirname(src_path or ""))


def normalize_requirements(text):
    """Normalize requirements wording in a string.

    Parameters
    ----------
    text : str

    Returns
    -------
    text : str
    """
    for level in ("optional", "recommended", "required", "deprecated"):
        # Replace both "optional" and "Optional" with "OPTIONAL"
        text = text.replace(level.title(), level).replace(level, level.upper())
    return text


def normalize_breaks(text):
    """Normalize line breaks in a string, for new lines, escaped new lines and double new lines.

    Parameters
    ----------
    text : str

    Returns
    -------
    text : str
    """
    # A backslash before a newline means continue a string
    text = text.replace("\\\n", "")
    # Two newlines should be respected
    text = text.replace("\n\n", "<br>")
    # Otherwise a newline corresponds to a space
    return text.replace("\n", " ")


def num2words(integer, to="ordinal"):
    """Convert integers to words.

    This is a very simplistic mapping of numbers to words,
    to avoid adding num2words to our requirements.
    It only works with the first few numbers.

    Parameters
    ----------
    integer : int
    to : {"ordinal", "cardinal"}, optional

    Returns
    -------
    word : str
    """
    if to == "ordinal":
        mapper = {
            1: "first",
            2: "second",
            3: "third",
            4: "fourth",
            5: "fifth",
            6: "sixth",
            7: "seventh",
            8: "eighth",
            9: "ninth",
            10: "tenth",
        }
    elif to == "cardinal":
        mapper = {
            1: "one",
            2: "two",
            3: "three",
            4: "four",
            5: "five",
            6: "six",
            7: "seven",
            8: "eight",
            9: "nine",
            10: "ten",
        }

    try:
        return mapper[integer]
    except KeyError:
        raise ValueError(f"Input {integer} is not supported.")



"""Utility functions for the bids-specification schema."""

import logging
import os
import sys

from . import data


def get_bundled_schema_path():
    """Get the path to the schema directory.

    Returns
    -------
    str
        Absolute path to the directory containing schema-related files.
    """
    return str(data.load_resource("schema"))


def get_logger(name=None, level=None):
    """Return a logger to use.

    Parameters
    ----------
    name : None or str, optional
        Name of the logger. Defaults to None.
    level: None or int, optional
        Level to set for the logger. Defaults to None.
        BIDS_SCHEMA_LOG_LEVEL environment variable can
        be used to set the level and will overwrite
        this value.

    Returns
    -------
    logging.Logger
        logger object.
    """
    logger = logging.getLogger("bidsschematools" + (f".{name}" if name else ""))
    # If explicitly instructed via env var -- set log level
    if log_level := os.getenv("BIDS_SCHEMA_LOG_LEVEL", level):
        set_logger_level(logger, log_level)
    return logger


def configure_logger(lgr):
    """Configuring formatting and stream handler for the logger.

    Should not be used when bidsschematools is used as a library.

    Parameters
    ----------
    lgr: logging.Logger

    """
    if len(lgr.handlers) == 0:
        # add a handler if there isn't one
        ch = logging.StreamHandler(stream=sys.stderr)
        lgr.addHandler(ch)
    # Set the formatter for the handlers
    for lh in lgr.handlers:
        lh.setFormatter(logging.Formatter("%(asctime)-15s [%(levelname)8s] %(message)s"))


def set_logger_level(lgr, level):
    """Set the logger level.

    Parameters
    ----------
    lgr : logging.Logger
        logger object for which to change the level.
    level : int or str
        The desired level for the logger.
    """
    if isinstance(level, int):
        pass
    elif level.isnumeric():
        level = int(level)
    elif level.isalpha():
        level = getattr(logging, level)
    else:
        lgr.warning("Do not know how to treat loglevel %s", level)
        return
    lgr.setLevel(level)



"""Utilities for implementing ``schema.rules``

This module is currently limited to constructing filename rules from
``schema.rules.files``.
"""

import fnmatch
import re
from collections.abc import Mapping
from functools import lru_cache

import bidsschematools as bst
import bidsschematools.types

# The list of which entities create directories could be dynamically specified by the YAML, but for
# now, it is not.
# Ordering is important, as "subject" follows "session" alphabetically, but is hierarchically
# above it.
DIR_ENTITIES = ["subject", "session"]


def _capture_regex(name, pattern, backref):
    """Capture pattern to name or match back-reference to name

    >>> _capture_regex("run", "[0-9]+", False)
    '(?P<run>[0-9]+)'
    >>> _capture_regex("run", "[0-9]+", True)
    '(?P=run)'
    >>> re.match(_capture_regex("run", "[0-9]+", False), "123_").groupdict()
    {'run': '123'}
    """
    return f"(?P={name})" if backref else f"(?P<{name}>{pattern})"


def _optional_regex(regex, optional):
    """Return an optional version of a regex if optional is True

    A required regex is passed through unchanged:

    >>> pattern = _optional_regex("xyz", False)
    >>> pattern
    'xyz'
    >>> re.match(pattern, "xyz").groups()
    ()
    >>> re.match(pattern, "") is None
    True

    An optional regex uses a non-capturing group, to avoid interfering
    with existing groups

    >>> pattern = _optional_regex("x(?P<name>[a-z])z", True)
    >>> pattern
    '(?:x(?P<name>[a-z])z)?'
    >>> re.match(pattern, "xyz").groups()
    ('y',)
    >>> re.match(pattern, "xyz").groupdict()
    {'name': 'y'}
    >>> re.match(pattern, "").groups()
    (None,)
    >>> re.match(pattern, "").groupdict()
    {'name': None}
    """
    return f"(?:{regex})?" if optional else regex


@lru_cache
def _format_entity(entity, name, pattern, level, directory=False):
    if directory and entity not in DIR_ENTITIES:
        return ""

    # For a directory entity appearing in a filename, match the value
    # found in the directory name
    # Note that this makes it impossible to do something like
    # /ses-1_dwi.json instead of /sub-??/ses-1/sub-??_ses-1_dwi.json
    # We'll assume this is a negligible use case.
    # It is possible to create such a regular expression, but doing it
    # in a piecewise fashion is difficult.
    if not directory and entity in DIR_ENTITIES:
        return rf"(?({entity}){name}-(?P={entity})_)"

    label = _capture_regex(entity, pattern, not directory and entity in DIR_ENTITIES)
    post = "/" if directory else "_"

    return _optional_regex(f"{name}-{label}{post}", level != "required")


def _entity_rule(rule: Mapping, schema: bst.types.Namespace):
    dir_regex = []
    entity_regex = []
    for ent in schema.rules.entities:
        if ent not in rule["entities"]:
            continue
        ent_obj = rule["entities"][ent]
        if isinstance(ent_obj, str):
            ent_obj = {"level": ent_obj}
        # Allow filename rule to override original entity fields
        entity = {**schema.objects.entities[ent], **ent_obj}

        if "enum" in entity:
            allowed_values = []
            for value in entity["enum"]:
                if isinstance(value, str):
                    allowed_values.append(value)
                else:
                    allowed_values.append(value["name"])

            pattern = "|".join(allowed_values)
        else:
            pattern = schema.objects.formats[entity["format"]].pattern

        dir_regex.append(
            _format_entity(ent, entity["name"], pattern, entity["level"], directory=True)
        )
        entity_regex.append(_format_entity(ent, entity["name"], pattern, entity["level"]))

    dtypes = set(rule.get("datatypes", ()))
    optional_dtype = "" in dtypes
    if optional_dtype:
        dtypes.remove("")
    if dtypes:
        pattern = f"(?P<datatype>{'|'.join(dtypes)})/"
        if optional_dtype:
            pattern = f"(?:{pattern})?"
        dir_regex += pattern

    # If we move to referring to suffixes by keys in the object table:
    # suffixes = [schema.objects.suffixes[suffix].value for suffix in rule["suffixes"]]
    suffixes = rule["suffixes"]
    suffix_regex = f"(?P<suffix>{'|'.join(suffixes)})"

    # If we move to referring to extensions by keys in the object table:
    # extensions = [schema.objects.extensions[ext].value for ext in rule["extensions"]]
    extensions = rule["extensions"]
    ext_match = "|".join(_sanitize_extension(ext) for ext in extensions)
    ext_regex = f"(?P<extension>{ext_match})"

    return {
        "regex": "".join(dir_regex + entity_regex + [suffix_regex, ext_regex, r"\Z"]),
        "mandatory": False,
    }


def _split_inheritance_rules(rule: dict) -> list[dict]:
    """Break composite rules into main and sidecar rules

    Implements the inheritance principle for file naming.
    """
    heritable_exts = {".tsv", ".json", ".bval", ".bvec"}
    rule_exts = set(rule["extensions"])

    main_exts = rule_exts - heritable_exts
    sidecar_exts = rule_exts - main_exts
    if not sidecar_exts:
        return [rule]

    rules = []

    # Some rules only address metadata, such as events.tsv or coordsystem.json
    if main_exts:
        rules.append({**rule, **{"extensions": list(main_exts)}})

    rules.append(
        {
            **rule,
            **{
                "extensions": list(sidecar_exts),
                "datatypes": [""] + rule.get("datatypes", []),
                "entities": {ent: "optional" for ent in rule["entities"]},
            },
        }
    )

    return rules


def _sanitize_extension(ext: str) -> str:
    if ext == ".*":
        return r"\.[a-zA-Z0-9.]+"
    return re.escape(ext)


def _stem_rule(rule: bst.types.Namespace):
    # translate includes a trailing \Z (end of string) but we expect extensions
    stem_match = fnmatch.translate(rule.stem)[:-2]
    stem_regex = f"(?P<stem>{stem_match})"

    dtypes = set(rule.get("datatypes", ()))
    dir_regex = f"(?P<datatype>{'|'.join(dtypes)})/" if dtypes else ""

    ext_match = "|".join(_sanitize_extension(ext) for ext in rule.extensions)
    ext_regex = rf"(?P<extension>{ext_match})\Z"

    return {"regex": dir_regex + stem_regex + ext_regex, "mandatory": rule.level == "required"}


def _path_rule(rule: bst.types.Namespace):
    path_match = re.escape(rule.path)
    # Exact path matches may be files or opaque directories
    # Consider using rules.directories to identify opaque directories
    return {"regex": rf"(?P<path>{path_match})(?:/.*)?\Z", "mandatory": rule.level == "required"}


def regexify_filename_rules(
    rule_group: bst.types.Namespace,
    schema: bst.types.Namespace,
    level: int,
):
    """Load schema rules into regular expressions

    Parameters
    ----------
    rule_group : Namespace
        The set of rules to load from the schema
    schema : Namespace
        A nested dictionary, as returned by `bidsschematools.schema.load_schema()`.
    level : int
        The depth in rule_group to look for rules

    Returns
    -------
    rules : list of dict
        A list of dictionaries, with keys including 'regex' and 'mandatory'.
    """
    regex_schema = []
    for rule_template in rule_group.values(level=level):
        # Simple rules, e.g. dataset_description.json, README
        if "path" in rule_template:
            regex_schema.append(_path_rule(rule_template))
        elif "stem" in rule_template:
            regex_schema.append(_stem_rule(rule_template))
        else:
            regex_schema.extend(
                _entity_rule(rule, schema) for rule in _split_inheritance_rules(rule_template)
            )

    return regex_schema


@lru_cache
def regexify_all(schema_dir=None):
    """
    Create full path regexes for all BIDS specification files.

    Parameters
    ----------
    schema_dir : str, optional
        A string pointing to a BIDS directory for which paths should be validated.

    Returns
    -------
    all_regex : list of dict
        A list of dictionaries, with keys including 'regex' and 'mandatory'.
    my_schema : Mapping
        Nested dictionaries representing the full schema.
    """

    schema = bst.schema.load_schema(schema_dir)
    all_regex = []
    for group in (schema.rules.files.common, schema.rules.files.raw):
        all_regex.extend(regexify_filename_rules(group, schema, level=2))

    return all_regex, schema



"""Functions for rendering portions of the schema as text."""

import yaml
from markdown_it import MarkdownIt

from bidsschematools.render import utils
from bidsschematools.schema import Namespace, filter_schema, load_schema
from bidsschematools.utils import get_logger

lgr = get_logger()

# Remember to add extension (.html or .md) to the paths when using them.
ENTITIES_PATH = "SPEC_ROOT/appendices/entities"
GLOSSARY_PATH = "SPEC_ROOT/glossary"
TYPE_CONVERTER = {
    "columns": "column",
    "common_principles": "common principle",
    "datatypes": "datatype",
    "entities": "entity",
    "extensions": "extension",
    "files": "files and directories",
    "formats": "format",
    "metadata": "metadata",
    "top_level_files": "top level file",
    "suffixes": "suffix",
}


def make_entity_definitions(schema, src_path=None):
    """Generate definitions and other relevant information for entities in the specification.

    Each entity gets its own heading.

    Parameters
    ----------
    schema : dict
        The schema object, which is a dictionary with nested dictionaries and
        lists stored within it.

    Returns
    -------
    text : str
        A string containing descriptions and some formatting
        information about the entities in the schema.
    """
    entity_order = schema["rules"]["entities"]
    entity_definitions = schema["objects"]["entities"]

    text = ""
    for entity in entity_order:
        entity_info = entity_definitions[entity]
        entity_text = _make_entity_definition(entity, entity_info)
        text += "\n" + entity_text

    text = text.replace("SPEC_ROOT", utils.get_relpath(src_path))
    return text


def _make_entity_definition(entity, entity_info):
    """Describe an entity."""
    entity_shorthand = entity_info["name"]
    text = ""
    text += f"## {entity_shorthand}"
    text += "\n\n"
    text += f"**Full name**: {entity_info['display_name']}"
    text += "\n\n"
    text += f"**Format**: `{entity_info['name']}-<{entity_info.get('format', 'label')}>`"
    text += "\n\n"
    if "enum" in entity_info.keys():
        allowed_values = []
        for value in entity_info["enum"]:
            if isinstance(value, str):
                allowed_values.append(value)
            else:
                allowed_values.append(value["name"])

        text += f"**Allowed values**: `{'`, `'.join(allowed_values)}`"
        text += "\n\n"

    description = entity_info["description"]
    text += f"**Definition**: {description}"
    return text


def make_glossary(schema, src_path=None):
    """Generate glossary.

    Parameters
    ----------
    schema : dict
        The schema object, which is a dictionary with nested dictionaries and
        lists stored within it.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.

    Returns
    -------
    text : str
        A string containing descriptions and some formatting
        information about the entities in the schema.
    """
    all_objects = {}
    schema = schema.to_dict()

    for group, group_objects in schema["objects"].items():
        group_obj_keys = list(group_objects.keys())

        # Do not include private objects in the glossary
        group_obj_keys = [k for k in group_obj_keys if not k.startswith("_")]

        # Identify multi-sense objects (multiple entries, indicated by __ in key)
        multi_sense_objects = []
        for key in group_obj_keys:
            if "__" in key:
                temp_key = key.split("__")[0]
                multi_sense_objects.append(temp_key)

        multi_sense_objects = sorted(list(set(multi_sense_objects)))
        sense_keys = {mso: [] for mso in multi_sense_objects}

        for key in group_obj_keys:
            for sense_key in sense_keys.keys():
                if (key == sense_key) or (key.startswith(sense_key + "__")):
                    sense_keys[sense_key].append(key)

        sense_names = {}
        for sense_key, key_list in sense_keys.items():
            for i_key, key in enumerate(key_list):
                new_key_name = f"{sense_key} _sense {i_key + 1}_"
                sense_names[key] = new_key_name

        for key in group_obj_keys:
            new_name = sense_names.get(key, key)
            new_name = f"{new_name} ({group})"
            all_objects[new_name] = {}
            all_objects[new_name]["key"] = f"objects.{group}.{key}"
            all_objects[new_name]["type"] = TYPE_CONVERTER.get(group, group)
            all_objects[new_name]["definition"] = group_objects[key]

    text = ""
    for obj_key in sorted(all_objects.keys()):
        obj = all_objects[obj_key]
        obj_marker = obj["key"]
        obj_def = obj.get("definition", None)
        if obj_def is None:
            raise ValueError(f"{obj_marker} has no definition.")

        # Clean up the text description
        obj_desc = obj_def.get("description", None)
        if obj_desc is None:
            raise ValueError(f"{obj_marker} has no description.")
        # A backslash before a newline means continue a string
        obj_desc = obj_desc.replace("\\\n", "")
        # Two newlines should be respected
        obj_desc = obj_desc.replace("\n\n", "<br>")
        # Otherwise a newline corresponds to a space
        obj_desc = obj_desc.replace("\n", " ")

        text += f'\n<a name="{obj_marker}"></a>'
        text += f"\n## {obj_key}\n\n"
        text += f"**Name**: {obj_def['display_name']}\n\n"
        text += f"**Type**: {obj['type'].title()}\n\n"

        if obj["type"] == "suffix":
            text += f"**Format**: `<entities>_{obj_def['value']}.<extension>`\n\n"
        elif obj["type"] == "extension":
            text += f"**Format**: `<entities>_<suffix>{obj_def['value']}`\n\n"
        elif obj["type"] == "format":
            text += f"**Regular expression**: `{obj_def['pattern']}`\n\n"

        keys_to_drop = [
            "description",
            "display_name",
            "name",
            "value",
            "pattern",
            "enum",
            "definition",
        ]
        levels = list(obj_def.get("enum", []) or obj_def.get("definition", {}).get("Levels", {}))
        if levels:
            levels = [level["name"] if isinstance(level, dict) else level for level in levels]
            text += f"**Allowed values**: `{'`, `'.join(levels)}`\n\n"

        text += f"**Description**:\n{obj_desc}\n\n"

        reduced_obj_def = {k: v for k, v in obj_def.items() if k not in keys_to_drop}

        if reduced_obj_def:
            reduced_obj_def = yaml.dump(reduced_obj_def)
            text += f"**Schema information**:\n```yaml\n{reduced_obj_def}\n```"

    # Spec internal links need to be replaced
    text = text.replace("SPEC_ROOT", utils.get_relpath(src_path))

    return text


def _add_entity(filename_template, entity_pattern, requirement_level):
    """Add entity pattern to filename template based on requirement level."""
    # Skip underscore on first entity
    if filename_template:
        entity_pattern = f"_{entity_pattern}"
    if requirement_level != "required":
        entity_pattern = f"[{entity_pattern}]"
    return filename_template + entity_pattern


def _format_entity(entity, lt, gt):
    fmt = entity.get("format")
    if "enum" in entity:
        allowed_values = []
        for value in entity["enum"]:
            if isinstance(value, str):
                allowed_values.append(value)
            else:
                allowed_values.append(value["name"])

        fmt = "|".join(allowed_values)

    if fmt is None:
        raise ValueError(f"entity missing format or enum fields: {entity}")

    return f"{entity['name']}-{lt}{fmt}{gt}"


def value_key_table(namespace):
    return {struct.value: key for key, struct in namespace.items()}


def make_filename_template(
    dstype,
    schema=None,
    src_path=None,
    n_dupes_to_combine=6,
    pdf_format=False,
    **kwargs,
):
    """Create codeblocks containing example filename patterns for a given datatype.

    By default, this function uses HTML, instead of direct Markdown codeblocks,
    so that it can embed hyperlinks within the filenames.

    Parameters
    ----------
    dstype : "raw" or "deriv"
        The type of files being rendered; determines if rules are found in rules.files.raw
        or rules.files.deriv
    schema : dict
        The schema object, which is a dictionary with nested dictionaries and
        lists stored within it.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.
    n_dupes_to_combine : int
        The minimum number of suffixes/extensions to combine in the template as
        <suffix>/<extension>.
    pdf_format : bool, optional
        If True, the filename template will be compiled as a standard markdown code block,
        without any hyperlinks, so that the specification's PDF build will look right.
        If False, the filename template will use HTML and include hyperlinks.
        This works on the website.
        Default is False.

    Other Parameters
    ----------------
    **kwargs : dict
        Keyword arguments used to filter the schema.
        Example kwargs that may be used include: "suffixes", "datatypes",
        "extensions".

    Returns
    -------
    codeblock : str
        A multiline string containing the filename templates for file types
        in the schema, after filtering.

    Notes
    -----
    This function links to HTML files, rather than markdown files.
    """
    if not schema:
        schema = load_schema()

    if pdf_format:
        lt, gt = "<", ">"
    else:
        lt, gt = "&lt;", "&gt;"

    schema = Namespace(filter_schema(schema.to_dict(), **kwargs))
    suffix_key_table = value_key_table(schema.objects.suffixes)
    ext_key_table = value_key_table(schema.objects.extensions)

    # Parent directories
    sub_string = utils._link_with_html(
        _format_entity(schema.objects.entities.subject, lt, gt),
        html_path=ENTITIES_PATH + ".html",
        heading="sub",
        pdf_format=pdf_format,
    )
    ses_string = utils._link_with_html(
        _format_entity(schema.objects.entities.session, lt, gt),
        html_path=ENTITIES_PATH + ".html",
        heading="ses",
        pdf_format=pdf_format,
    )
    lines = [f"{sub_string}/", f"\t[{ses_string}/]"]

    file_rules = schema.rules.files[dstype]
    file_groups = {}
    for rule in file_rules.values(level=2):
        for datatype in rule.datatypes:
            file_groups.setdefault(datatype, []).append(rule)

    for datatype in sorted(file_groups):
        datatype_string = utils._link_with_html(
            datatype,
            html_path=GLOSSARY_PATH + ".html",
            heading=f"{datatype.lower()}-datatypes",
            pdf_format=pdf_format,
        )
        lines.append(f"\t\t{datatype_string}/")

        # Unique filename patterns
        for group in file_groups[datatype]:
            ent_string = ""
            for ent in schema.rules.entities:
                if ent not in group.entities:
                    continue

                # Add level and any overrides to entity
                ent_obj = group.entities[ent]
                if isinstance(ent_obj, str):
                    ent_obj = {"level": ent_obj}
                entity = {**schema.objects.entities[ent], **ent_obj}

                if "enum" in entity:
                    # Link full entity
                    pattern = utils._link_with_html(
                        _format_entity(entity, lt, gt),
                        html_path=f"{ENTITIES_PATH}.html",
                        heading=entity["name"],
                        pdf_format=pdf_format,
                    )
                else:
                    # Link entity and format separately
                    entity["name"] = utils._link_with_html(
                        entity["name"],
                        html_path=f"{ENTITIES_PATH}.html",
                        heading=entity["name"],
                        pdf_format=pdf_format,
                    )
                    fmt = entity.get("format", "label")
                    entity["format"] = utils._link_with_html(
                        entity.get("format", "label"),
                        html_path=f"{GLOSSARY_PATH}.html",
                        heading=f"{fmt}-common_principles",
                        pdf_format=pdf_format,
                    )
                    pattern = _format_entity(entity, lt, gt)

                ent_string = _add_entity(ent_string, pattern, entity["level"])

            # In cases of large numbers of suffixes,
            # we use the "suffix" variable and expect a table later in the spec
            if len(group["suffixes"]) >= n_dupes_to_combine:
                suffixes = [
                    lt
                    + utils._link_with_html(
                        "suffix",
                        html_path=GLOSSARY_PATH + ".html",
                        heading="suffix-common_principles",
                        pdf_format=pdf_format,
                    )
                    + gt
                ]
            else:
                suffixes = [
                    utils._link_with_html(
                        suffix,
                        html_path=GLOSSARY_PATH + ".html",
                        heading=f"{suffix_key_table[suffix].lower()}-suffixes",
                        pdf_format=pdf_format,
                    )
                    for suffix in group.suffixes
                ]

            # Add extensions
            extensions = [ext if ext != "*" else ".<extension>" for ext in group.extensions]
            if len(extensions) >= n_dupes_to_combine:
                # Combine exts when there are many, but keep JSON separate
                if ".json" in extensions:
                    extensions = [".<extension>", ".json"]
                else:
                    extensions = [".<extension>"]

            ext_headings = []
            for extension in extensions:
                # The glossary indexes by the extension identifier (nii_gz instead of .nii.gz),
                # but the rules reference the actual suffix string (.nii.gz instead of nii_gz),
                # so we need to look it up.
                key = ext_key_table.get(extension)
                if key:
                    ext_headings.append(f"{key.lower()}-extensions")
                else:
                    ext_headings.append("extension-common_principles")

            extensions = utils.combine_extensions(
                extensions,
                html_path=GLOSSARY_PATH + ".html",
                heading_lst=ext_headings,
                pdf_format=pdf_format,
            )

            lines.extend(
                f"\t\t\t{ent_string}_{suffix}{extension}"
                for suffix in sorted(suffixes)
                for extension in sorted(extensions)
            )

    paragraph = "\n".join(lines)
    if pdf_format:
        codeblock = f"Template:\n```Text\n{paragraph}\n```"
    else:
        codeblock = (
            f'Template:\n<div class="highlight"><pre><code>{paragraph}\n</code></pre></div>'
        )

    codeblock = codeblock.expandtabs(4)
    codeblock = append_filename_template_legend(codeblock, pdf_format)
    codeblock = codeblock.replace("SPEC_ROOT", utils.get_relpath(src_path))

    return codeblock


def append_filename_template_legend(text, pdf_format=False):
    """Append a legend to filename templates.

    Parameters
    ----------
    text : str
        The text to append the legend to.

    pdf_format : bool
        Whether to format the legend for PDF output.

    Returns
    -------
    str :
        The text with the legend appended.
    """
    if pdf_format:
        info_str = ""
    else:
        info_str = """
- For more information about filename elements (for example, entities, suffixes, extensions),
  follow the links embedded in the filename template.
  """

    legend = f"""{info_str}
- Filename entities or directories between square brackets
  (for example, `[_ses-<label>]`) are OPTIONAL.
- Some entities may only allow specific values,
  in which case those values are listed in `<>`, separated by `|`.
- `_<suffix>` means that there are several (>6) valid suffixes for this filename pattern.
- `.<extension>` means that there are several (>6) valid extensions for this file type.
- `[.gz]` means that both the unzipped and gzipped versions of the extension are valid.
"""

    if pdf_format:
        text += f"""

**Legend**:

{legend}

"""
    else:
        md = MarkdownIt()
        text += f"""
<details>
<summary><strong>Legend:</strong></summary>
{md.render(legend)}
</details>
"""

    return text


def define_common_principles(schema, src_path=None):
    """Enumerate the common principles defined in the schema.

    Parameters
    ----------
    schema : dict
        The BIDS schema.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.

    Returns
    -------
    string : str
        The definitions of the common principles in a multiline string.
    """
    string = ""
    common_principles = schema["objects"]["common_principles"]
    order = schema["rules"]["common_principles"]
    for i_prin, principle in enumerate(order):
        principle_name = common_principles[principle]["display_name"]
        substring = (
            f"{i_prin + 1}. **{principle_name}** - {common_principles[principle]['description']}"
        )
        string += substring
        if i_prin < len(order) - 1:
            string += "\n\n"

    string = string.replace("SPEC_ROOT", utils.get_relpath(src_path))

    return string


def define_allowed_top_directories(schema, src_path=None) -> str:
    """Create a list of allowed top-level directories with their descriptions.

    Parameters
    ----------
    schema : dict
        The BIDS schema.
    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.

    Returns
    -------
    string : str
        Unordered list describing top level directories.
    """

    string = ""

    for dirname, definition in schema.objects.files.items():
        if definition.file_type == "directory":
            string += f"- `{dirname}`: {definition.description}"

    return string.replace("SPEC_ROOT", utils.get_relpath(src_path))


def render_text(schema, key: str, src_path=None):
    """

    Parameters
    ----------
    schema : dict
        The BIDS schema.

    key : str
        The key of the object to render the description for:
        possible values correspond to the keys in schema["objects"][object]

    src_path : str or None
        The file where this macro is called, which may be explicitly provided
        by the "page.file.src_path" variable.

    Returns
    -------
    desc : str
        Description of the object.
    """
    text = schema.get(key)
    if not isinstance(text, str):
        raise ValueError(f"{key} does not refer to a text field")
    return text.replace("SPEC_ROOT", utils.get_relpath(src_path))



"""Tests for the bidsschematools package."""

import os
from collections.abc import Mapping

import pytest
from jsonschema.exceptions import ValidationError

from bidsschematools import __bids_version__, schema, types


def test__get_bids_version(schema_dir):
    # Is the version being read in correctly?
    bids_version = schema._get_bids_version(schema_dir)
    assert bids_version == __bids_version__


def test__get_bids_version_fallback(tmp_path):
    # Does fallback to unknown development version work?
    expected_version = "1.2.3-dev"
    schema_path = os.path.join(tmp_path, "whatever", expected_version)
    bids_version = schema._get_bids_version(schema_path)
    assert bids_version == expected_version

    # Does fallback to path quoting work?
    schema_path = os.path.join(tmp_path, "whatever", "undocumented_schema_dir")
    bids_version = schema._get_bids_version(schema_path)
    assert bids_version == schema_path


def test_load_schema(schema_dir):
    """Smoke test for bidsschematools.schema.load_schema."""
    # Pointing to a nonexistent directory should raise a ValueError
    bad_path = "/path/to/nowhere"
    with pytest.raises(FileNotFoundError):
        schema.load_schema(bad_path)

    # Otherwise the function should return a dictionary
    schema_obj = schema.load_schema(schema_dir)
    assert isinstance(schema_obj, Mapping)

    # Check that it is fully dereferenced
    assert "$ref" not in str(schema_obj)


def test_object_definitions(schema_obj):
    """Ensure that object definitions in the schema contain required fields."""
    for obj_type, obj_type_def in schema_obj["objects"].items():
        for obj_key, obj_def in obj_type_def.items():
            # Private/inheritable definitions (ones starting with "_") do not need to conform to
            # the same rules as user-facing terms, so we skip them
            if obj_key.startswith("_"):
                continue

            assert "display_name" in obj_def, obj_key
            assert "description" in obj_def.keys(), obj_key
            if obj_type in ("columns", "entities", "metadata"):
                assert "name" in obj_def
            elif obj_type in ("datatypes", "extensions", "suffixes"):
                assert "value" in obj_def


def test_formats(schema_obj):
    """Test valid string patterns allowed by the specification."""
    import re

    # Check that valid strings match the search pattern.
    GOOD_PATTERNS = {
        "label": ["01", "test", "test01", "Test01"],
        "index": ["01", "1", "10000", "00001"],
        "string": ["any string is valid."],
        "integer": ["5", "10", "-5", "-10"],
        "number": [
            "5",  # integers are allowed
            "3.14",  # floats too
            "-5",  # they can be negative
            "-3.14",
            "1e3",  # scientific notation is allowed
            "-2.1E+5",
        ],
        "boolean": ["true", "false"],
        "date": ["2022-01-05", "2022-01-05UTC", "2022-50-50"],
        "datetime": [
            "2022-01-05T13:16:30",
            "2022-01-05T13:16:30.5",  # subsecond resolution is allowed
            "2022-01-05T13:16:30.000005",  # up to 6 decimal points
            "2022-01-05T13:16:30Z",  # UTC indicator is allowed
            "2022-01-05T13:16:30.05Z",
        ],
        "time": [
            "13:16:30",
            "09:00:00",
            "9:00:00",  # leading zeros are not required for hours
        ],
        "unit": ["any string is valid."],
        "file_relative": [
            "file_in_same_directory.txt",
            "../../relative/path/file.txt",
            "sub-01/path/file.txt",
        ],
        "stimuli_relative": ["any/arbitrary/path/file.txt"],
        "dataset_relative": ["any/arbitrary/path/file.txt"],
        "participant_relative": ["any/arbitrary/path/file.txt"],
        "rrid": ["RRID:SCR_017398"],
        "uri": ["foo://example.com:8042/over/there?name=ferret#nose"],
        "bids_uri": [
            "bids::sub-01/fmap/sub-01_dir-AP_epi.nii.gz",
            "bids:ds000001:sub-02/anat/sub-02_T1w.nii.gz",
            "bids:myderivatives:sub-03/func/sub-03_task-rest_space-MNI152_bold.nii.gz",
        ],
    }
    for pattern, test_list in GOOD_PATTERNS.items():
        pattern_format = schema_obj["objects"]["formats"][pattern]["pattern"]
        search_pattern = "^" + pattern_format + "$"
        search = re.compile(search_pattern)
        for test_string in test_list:
            assert bool(search.fullmatch(test_string)), (
                f"'{test_string}' is not a valid match for the pattern '{search.pattern}'"
            )

    # Check that invalid strings do not match the search pattern.
    BAD_PATTERNS = {
        "label": ["test_01", "!", "010101-", "01-01", "-01"],
        "index": ["test", "0.1", "0-1", "0_1"],
        "string": [],
        "integer": ["3.14", "-3.14", "1.", "-1.", "string", "s1", "1%", "one"],
        "number": ["string", "1%"],
        "boolean": ["True", "False", "T", "F"],
        "date": [
            "05-01-2022",  # MM-DD-YYYY or DD-MM-YYYY
            "05/01/2022",  # MM/DD/YYYY or DD/MM/YYYY
        ],
        "datetime": [
            "2022-01-05T13:16:30.1000005",  # too many decimal points
            "2022-01-05T13:16:30U",  # Only Z is permitted
            "2022-01-05T13:16:30UTCUTC",  # time zone too long
            "2022-01-05T34:10:10",  # invalid time
        ],
        "time": [
            "34:10:10",  # invalid time
            "24:00:00",  # should be 00:00:00
            "00:60:00",  # should be 01:00:00
            "00:00:60",  # should be 00:01:00
            "01:23",  # lacks either hours or seconds
        ],
        "unit": [],
        "file_relative": [
            "/path/with/starting/slash/file.txt",
        ],
        "stimuli_relative": [
            "/path/with/starting/slash/file.txt",
            "stimuli/path/file.txt",
        ],
        "dataset_relative": [
            "/path/with/starting/slash/file.txt",
        ],
        "participant_relative": [
            "/path/with/starting/slash/file.txt",
            "sub-01/path/file.txt",
        ],
        "rrid": [
            "RRID:",  # empty one
        ],
        "uri": [
            # "ftp://",  # lacks anything but protocol. This should fail, but doesn't ATM.
        ],
        "bids_uri": [],
    }
    for pattern, test_list in BAD_PATTERNS.items():
        pattern_format = schema_obj["objects"]["formats"][pattern]["pattern"]
        search_pattern = f"^{pattern_format}$"
        search = re.compile(search_pattern)
        for test_string in test_list:
            assert not bool(search.fullmatch(test_string)), (
                f"'{test_string}' should not be a valid match for the pattern '{search.pattern}'"
            )


def test_dereferencing():
    orig = {
        "ReferencedObject": {
            "Property1": "value1",
            "Property2": "value2",
        },
        "ReferencingObject": {
            "$ref": "ReferencedObject",
            "Property2": "value4",
        },
    }
    dereffed = schema.dereference(orig)
    assert dereffed == {
        "ReferencedObject": {
            "Property1": "value1",
            "Property2": "value2",
        },
        "ReferencingObject": {
            "Property1": "value1",
            "Property2": "value4",
        },
    }

    orig = {
        "raw.func": {
            "suffix": ["bold", "cbv"],
            "extensions": [".nii", ".nii.gz"],
            "datatype": ["func"],
            "entities": {
                "subject": "required",
                "session": "optional",
                "task": "required",
                "dir": "optional",
            },
        },
        "derived.func": {
            "$ref": "raw.func",
            "entities": {
                "$ref": "raw.func.entities",
                "space": "optional",
                "desc": "optional",
            },
        },
    }

    sch = types.Namespace.build(orig)
    dereffed = schema.dereference(sch)
    assert dereffed == {
        "raw": {
            "func": {
                "suffix": ["bold", "cbv"],
                "extensions": [".nii", ".nii.gz"],
                "datatype": ["func"],
                "entities": {
                    "subject": "required",
                    "session": "optional",
                    "task": "required",
                    "dir": "optional",
                },
            }
        },
        "derived": {
            "func": {
                "suffix": ["bold", "cbv"],
                "extensions": [".nii", ".nii.gz"],
                "datatype": ["func"],
                "entities": {
                    "subject": "required",
                    "session": "optional",
                    "task": "required",
                    "dir": "optional",
                    "space": "optional",
                    "desc": "optional",
                },
            }
        },
    }

    orig = {
        "_DERIV_ENTS": {
            "space": "optional",
            "desc": "optional",
        },
        "raw.func": {
            "suffix": ["bold", "cbv"],
            "extensions": [".nii", ".nii.gz"],
            "datatype": ["func"],
            "entities": {
                "subject": "required",
                "session": "optional",
                "task": "required",
                "dir": "optional",
            },
        },
        "derived.func": {
            "$ref": "raw.func",
            "entities": {
                "$ref": "_DERIV_ENTS",
            },
        },
    }

    sch = types.Namespace.build(orig)
    dereffed = schema.dereference(sch)
    assert dereffed == {
        "_DERIV_ENTS": {
            "space": "optional",
            "desc": "optional",
        },
        "raw": {
            "func": {
                "suffix": ["bold", "cbv"],
                "extensions": [".nii", ".nii.gz"],
                "datatype": ["func"],
                "entities": {
                    "subject": "required",
                    "session": "optional",
                    "task": "required",
                    "dir": "optional",
                },
            }
        },
        "derived": {
            "func": {
                "suffix": ["bold", "cbv"],
                "extensions": [".nii", ".nii.gz"],
                "datatype": ["func"],
                "entities": {
                    "space": "optional",
                    "desc": "optional",
                },
            }
        },
    }

    orig = {
        "objects": {
            "enums": {
                "left": {"value": "L"},
                "right": {"value": "R"},
            },
            "entities.hemisphere": {
                "name": "hemi",
                "enum": [
                    {"$ref": "objects.enums.left.value"},
                    {"$ref": "objects.enums.right.value"},
                ],
            },
        },
    }

    sch = types.Namespace.build(orig)
    dereffed = schema.dereference(sch)
    assert dereffed == {
        "objects": {
            "enums": {
                "left": {"value": "L"},
                "right": {"value": "R"},
            },
            "entities": {
                "hemisphere": {
                    "name": "hemi",
                    "enum": ["L", "R"],
                },
            },
        },
    }


def test_namespace_to_dict():
    def check_for_namespaces(obj):
        if isinstance(obj, dict):
            [check_for_namespaces(val) for val in obj.values()]
        elif isinstance(obj, list):
            [check_for_namespaces(val) for val in obj]
        elif isinstance(obj, types.Namespace):
            raise ValueError("Namespace object found in dict")

    check_for_namespaces(schema.load_schema().to_dict())


def test_valid_schema():
    """Test that a valid schema does not raise an error."""
    namespace = schema.load_schema()
    schema.validate_schema(namespace)


def test_add_legal_field():
    """Test that adding a legal field does not raise an error."""
    namespace = schema.load_schema()
    namespace["rules"]["files"]["deriv"]["preprocessed_data"]["anat_nonparametric_common"][
        "entities"
    ]["density"] = "optional"
    schema.validate_schema(namespace)


def test_invalid_value():
    """Test that an invalid value raises an error."""
    namespace = schema.load_schema()
    namespace["rules"]["files"]["deriv"]["preprocessed_data"]["anat_nonparametric_common"][
        "entities"
    ]["density"] = "invalid"
    with pytest.raises(ValidationError) as e:
        schema.validate_schema(namespace)
    print(e.value)
    assert "invalid" in str(e.value)



import sys

sys.path.insert(0, "src")

import bidsschematools.schema
from bidsschematools.types._generator import generate_module


def pdm_build_initialize(context):
    context.ensure_build_dir()

    # Retrieve package version from schema
    schema = bidsschematools.schema.load_schema()
    context.config.metadata["version"] = schema.schema_version

    if context.target == "editable":
        return

    # src/ layout makes sdists different from wheels
    base_dir = context.build_dir / ("src" if context.target == "sdist" else "")

    # Write compiled schema to JSON
    schema_json = base_dir / "bidsschematools/data/schema.json"
    schema_json.parent.mkdir(parents=True, exist_ok=True)
    schema_json.write_text(schema.to_json())

    # Write generated code for types
    # Limit to wheel to avoid duplication while allowing building
    # the wheel directly from source
    if context.target == "wheel":
        context_py = base_dir / "bidsschematools/types/context.py"
        context_py.parent.mkdir(parents=True, exist_ok=True)
        context_py.write_text(generate_module(schema, "dataclasses"))

        protocols_py = base_dir / "bidsschematools/types/protocols.py"
        protocols_py.parent.mkdir(parents=True, exist_ok=True)
        protocols_py.write_text(generate_module(schema, "protocol"))


def pdm_build_update_files(context, files):
    # Dereference symlinks
    files.update({relpath: path.resolve() for relpath, path in files.items()})
    # Remove code generator, which is not used once installed
    if context.target == "wheel":
        del files["bidsschematools/types/_generator.py"]



from bidsschematools.render import utils


def test_combine_extensions():
    """A unit test for utils.combine_extensions."""
    test_extensions = ["nii.gz", "nii", "json"]
    target_combined = ["nii[.gz]", "json"]
    test_combined = utils.combine_extensions(test_extensions, pdf_format=True)
    assert test_combined == target_combined


def test_resolve_metadata_type():
    """A unit test for utils.resolve_metadata_type."""
    base_definition = {
        "name": "Term",
        "description": "A description",
    }

    # Basic string
    term_definition1 = base_definition.copy()
    term_definition1["type"] = "string"
    target_description = "[string](https://www.w3schools.com/js/js_json_datatypes.asp)"
    type_description = utils.resolve_metadata_type(term_definition1)
    assert target_description == type_description

    # When n/a is the only allowed value, the type should say "n/a"
    term_definition1["enum"] = ["n/a"]
    target_description = '`"n/a"`'
    type_description = utils.resolve_metadata_type(term_definition1)
    assert target_description == type_description
